import json
import os
import tempfile
from unittest.mock import Mock, patch

import pytest

from energinetml.core.model import (
    LoadedModel,
    MetaFileMissing,
    Model,
    ModelArtifact,
    ModelModuleError,
    TrainedModel,
    import_model_class,
)
from energinetml.core.project import MachineLearningProject
from energinetml.core.requirements import RequirementList
from energinetml.settings import DEFAULT_RELATIVE_ARTIFACT_PATH
from tests.conftest import create_model
from tests.constants import (
    COMPUTE_TARGET,
    DATASETS,
    EXPERIMENT,
    EXPERIMENT_NAME,
    FEATURES,
    FILES_INCLUDE,
    MODEL_NAME,
    PARAMETERS,
    RUN_ID,
    VM_SIZE,
)


class TestModel:
    def test__project__project_does_not_exists__should_return_none(self, model):
        """
        :param Model model:
        """
        assert model.project is None

    def test__project__project_exists__should_return_project_object(
        self, model_with_project
    ):  # noqa: E501
        """
        :param Model model_with_project:
        """
        assert isinstance(model_with_project.project, MachineLearningProject), type(
            model_with_project.project
        )

    def test__trained_model_path(self, model):
        """
        :param Model model:
        """
        assert model.trained_model_path == os.path.join(
            model.path, DEFAULT_RELATIVE_ARTIFACT_PATH, Model._TRAINED_MODEL_FILE_NAME
        )

    def test__data_folder_path(self, model):
        """
        :param Model model:
        """
        assert model.data_folder_path == os.path.join(model.path, "data")

    def test__requirements_file_path(self, model):
        """
        :param Model model:
        """
        assert model.requirements_file_path == os.path.join(
            model.path, "requirements.txt"
        )

    def test__datasets_parsed(self, model):
        """
        :param Model model:
        """
        datasets_local = dict(model.datasets_parsed.local)
        datasets_cloud = dict(model.datasets_parsed.cloud)

        assert datasets_local["iris"] is None
        assert datasets_local["local"] == "2"

        assert datasets_cloud["iris"] is None
        assert datasets_cloud["cloud"] == "3"

    def test__requirements__model_has_requirements_file__should_return_model_requirements(  # noqa: E501
        self, model
    ):
        """
        :param Model model:
        """
        with open(os.path.join(model.path, "requirements.txt"), "w") as f:
            f.write("some-package==1.0.0")

        assert isinstance(model.requirements, RequirementList)
        assert "some-package" in model.requirements
        assert model.requirements.get("some-package").specs == [("==", "1.0.0")]

    def test__requirements__model_has_no_requirements_file_but_project_has__should_return_project_requirements(  # noqa: E501
        self, model_with_project
    ):
        """
        :param Model model_with_project:
        """
        assert isinstance(model_with_project.requirements, RequirementList)
        assert (
            model_with_project.requirements is model_with_project.project.requirements
        )

    def test__requirements__no_requirements_exists__should_return_empty_requirements_list(  # noqa: E501
        self, model
    ):
        """
        :param Model model:
        """
        assert isinstance(model.requirements, RequirementList)
        assert model.requirements == []

    @patch("energinetml.core.model.FileMatcher")
    def test__files__should_return_file_matcher_object(self, file_matcher_class, model):
        """
        :param Mock file_matcher_class:
        :param Model model:
        """
        file_matcher_instance = Mock()
        file_matcher_class.return_value = file_matcher_instance

        assert model.files is file_matcher_instance
        file_matcher_class.assert_called_once_with(
            root_path=model.path,
            include=model.files_include,
            exclude=model.files_exclude,
            recursive=True,
        )

    def test__generate_seed(self, model):
        """
        :param Model model:
        """
        assert isinstance(model.generate_seed(), int)
        assert 0 <= model.generate_seed() <= 10**9

    def test__train(self, model):
        """
        :param Model model:
        """
        with pytest.raises(NotImplementedError):
            model.train(datasets=[], logger=None, seed=None)

    def test__predict(self, model):
        """
        :param Model model:
        """
        with pytest.raises(NotImplementedError):
            model.predict(trained_model=None, identifier=None, input_data=None)

    def test__create__should_create_project_files(self):
        with tempfile.TemporaryDirectory() as path:
            project = create_model(path)

            assert os.path.isfile(os.path.join(project.path, "model.json"))
            assert os.path.isfile(os.path.join(project.path, "model.py"))
            assert os.path.isfile(os.path.join(project.path, "__init__.py"))

            # model.json
            with open(os.path.join(project.path, "model.json")) as f:
                config = json.load(f)
                assert config["name"] == MODEL_NAME
                assert config["experiment"] == EXPERIMENT
                assert config["compute_target"] == COMPUTE_TARGET
                assert config["vm_size"] == VM_SIZE
                assert config["datasets"] == DATASETS
                assert config["features"] == FEATURES
                assert config["parameters"] == PARAMETERS

    def test__as_dict__should_return_object_as_a_dictionary(self, model_with_project):
        # Arrange in fixture

        # Act
        return_dict = model_with_project.as_dict()

        # Assert
        assert type(return_dict) == dict
        assert return_dict["name"] == MODEL_NAME
        assert return_dict["experiment"] == EXPERIMENT
        assert return_dict["compute_target"] == COMPUTE_TARGET
        assert return_dict["vm_size"] == VM_SIZE
        assert return_dict["datasets"] == DATASETS
        assert return_dict["features"] == FEATURES
        assert return_dict["parameters"] == PARAMETERS

    def test__load__should_return_loaded_model(self, model_path):

        # Arrange (also from fixture in parameter)
        trained_model_path = os.path.join(
            model_path,
            DEFAULT_RELATIVE_ARTIFACT_PATH,
            Model._TRAINED_MODEL_FILE_NAME,
        )

        # Act
        trained_model = Model.load(trained_model_path)

        # Assert
        assert type(trained_model) == TrainedModel

    def test__parent_levels__no_extra_levels__should_return_one(self):

        # Arrange
        model = Model(
            path=None,
            name=MODEL_NAME,
            experiment=EXPERIMENT,
            compute_target=COMPUTE_TARGET,
            vm_size=VM_SIZE,
            files_include=FILES_INCLUDE,
        )

        # Act
        parent_levels = model.parent_levels

        # Assert
        assert parent_levels == 1

    def test__parent_levels__one_extra_level__should_return_two(self):

        # Arrange
        files_include_with_parent_file = FILES_INCLUDE + ["../test.py"]
        model = Model(
            path=None,
            name=MODEL_NAME,
            experiment=EXPERIMENT,
            compute_target=COMPUTE_TARGET,
            vm_size=VM_SIZE,
            files_include=files_include_with_parent_file,
        )

        # Act
        parent_levels = model.parent_levels

        # Assert
        assert parent_levels == 2


class TestImportModelClass:
    def test__import_model_class__extra_parent_level__should_import_module_correct(
        self,
    ):
        # Arrange
        files_include_with_parent_file = FILES_INCLUDE + ["../test.py"]
        with tempfile.TemporaryDirectory() as path:

            model = create_model(path, files_include=files_include_with_parent_file)

            with open(os.path.join(model.path, "..", "test.py"), "w") as file:
                file.write("TEST_CONSTANT = 2")

            model_py_path = os.path.join(model.path, Model._SCRIPT_FILE_NAME)
            with open(model_py_path, "r") as original:
                data = original.read()
            with open(model_py_path, "w") as modified:
                modified.write("from ..test import TEST_CONSTANT\n" + data)

            # Act
            model_class = import_model_class(model.path, model.parent_levels)

            # Assert
            assert issubclass(model_class, Model)

    def test__import_model_class__include_submodule__should_import_module_correct(
        self,
    ):

        # Arrange
        files_include_with_parent_file = FILES_INCLUDE + ["test.py"]
        with tempfile.TemporaryDirectory() as path:

            model = create_model(path, files_include=files_include_with_parent_file)

            with open(os.path.join(model.path, "test.py"), "w") as file:
                file.write("TEST_CONSTANT = 2")

            model_py_path = os.path.join(model.path, Model._SCRIPT_FILE_NAME)
            with open(model_py_path, "r") as original:
                data = original.read()
            with open(model_py_path, "w") as modified:
                modified.write("from .test import TEST_CONSTANT\n" + data)

            # Act
            model_class = import_model_class(model.path, model.parent_levels)

            # Assert
            assert issubclass(model_class, Model)

    def test__import_from_model__submodule_not_in_files_include__should_raise_error(
        self,
    ):

        with tempfile.TemporaryDirectory() as path:

            # Setup model without specifying the "../test.py"
            model = create_model(path)
            dir_path = model.path + "/../shared"
            os.makedirs(dir_path)
            with open(os.path.join(dir_path, "utils.py"), "w") as file:
                file.write("TEST_CONSTANT = 2")

            model_py_path = os.path.join(model.path, Model._SCRIPT_FILE_NAME)
            with open(model_py_path, "r") as original:
                data = original.read()

            with open(model_py_path, "w") as modified:
                modified.write("from ...shared.utils import TEST_CONSTANT\n" + data)

            # Act and assert
            with pytest.raises(ModelModuleError):
                import_model_class(model.path, model.parent_levels)


class TestTrainedModel:
    def test__init_with_both_model_and_models__should_raise_value_error(self):
        with pytest.raises(ValueError):
            TrainedModel(model="x", models={"x": "y"})

    def test__identifiers(self):
        uut = TrainedModel(models={"x": "y", "z": "w"})

        assert uut.identifiers == ["x", "z"]

    def test__has_model(self):
        uut = TrainedModel(models={"x": "y", "z": "w"})

        assert uut.has_model("x")
        assert uut.has_model("z")
        assert not uut.has_model("y")

    def test__get_model__has_default_model(self):
        model = Mock()
        uut = TrainedModel(model=model)

        assert uut.get_model() is model

        with pytest.raises(ValueError):
            uut.get_model("x")

    def test__get_model__has_models(self):
        x = Mock()
        z = Mock()
        uut = TrainedModel(models={"x": x, "z": z})

        assert uut.get_model("x") is x
        assert uut.get_model("z") is z

        with pytest.raises(ValueError):
            uut.get_model()
        with pytest.raises(ValueError):
            uut.get_model("y")

    def test__get_default_model__has_default_model(self):
        model = Mock()
        uut = TrainedModel(model=model)

        assert uut.get_default_model() is model

    def test__get_default_model__has_models(self):
        x = Mock()
        z = Mock()
        uut = TrainedModel(models={"x": x, "z": z})

        with pytest.raises(ValueError):
            uut.get_default_model()

    def test__verify__features_is_not_a_list(self):
        uut = TrainedModel(model="x", features=None)

        with pytest.raises(uut.Invalid):
            uut.verify()

    def test__verify__features_is_not_strings_exclusively(self):
        uut = TrainedModel(model="x", features=["a", "b", 1])

        with pytest.raises(uut.Invalid):
            uut.verify()

    def test__verify__features_are_invalid(self):
        uut = TrainedModel(model="x", features=[" ", " "])

        with pytest.raises(uut.Invalid):
            uut.verify()

    def test__verify__features_are_valid__should_not_raise(self):
        uut = TrainedModel(model="x", features=["a", "b"])

        uut.verify()


class TestLoadedModel:
    def test__from_artifact__no_meta_data_file__should_raise_meta_file_missing(
        self, model_path
    ):

        model_artifact_mock = Mock()
        model_artifact_mock.path = model_path
        model_artifact_mock.run_id = RUN_ID
        model_artifact_mock.experiment_name = EXPERIMENT_NAME
        model_artifact_mock.portal_url = "URL"

        with pytest.raises(MetaFileMissing):
            LoadedModel.from_artifact(model_artifact_mock)

    @patch("energinetml.core.model.Model.load")
    @patch("energinetml.core.model.LoadedModel.read_meta_data")
    def test__from_artifact__should_return_loaded_model(
        self, read_meta_data_mock, model_load_mock, user_model
    ):
        # Arrange
        model_artifact = ModelArtifact(
            path=user_model.path,
            run_id=RUN_ID,
            experiment_name=EXPERIMENT_NAME,
            project_meta=None,
        )

        meta_data = {"module_name": user_model.module_name, "run_id": RUN_ID}
        read_meta_data_mock.return_value = meta_data
        trained_model = TrainedModel(
            model="123", params={"asd": 123}, features=FEATURES
        )
        model_load_mock.return_value = trained_model

        # Act
        loaded_model = LoadedModel.from_artifact(model_artifact)

        # Assert
        assert type(loaded_model) == LoadedModel

    def test__read_meta_data__should_read_correct(self):

        # Arrange
        model_artifact = ModelArtifact(
            path=None,
            run_id=RUN_ID,
            experiment_name=EXPERIMENT_NAME,
            project_meta=None,
        )

        meta_data_write = {"module_name": MODEL_NAME, "run_id": RUN_ID}
        with tempfile.TemporaryDirectory() as path:
            meta_file_path = os.path.join(path, Model._META_FILE_NAME)
            with open(meta_file_path, "w") as meta_data_file:
                json.dump(meta_data_write, meta_data_file)

            # Act
            meta_data_read = LoadedModel.read_meta_data(meta_file_path, model_artifact)

        # Assert
        assert meta_data_read == meta_data_write



"""This is the model module.

This module implements and exposes a Model class which defines the training logic and
inference logic.
"""

from .model import NewModel

# Reference your model class and name it "model"
model = NewModel



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List

from energinetml.core.submitting import SubmitContext
from energinetml.core.training import AbstractTrainingContext

if TYPE_CHECKING:
    from energinetml.core.model import Model


class ComputeType(Enum):
    """[summary]"""

    CPU = "CPU"
    GPU = "GPU"


class BackendException(Exception):
    """[summary]"""

    pass


class AbstractBackend:
    """[summary]"""

    BackendException = BackendException

    def get_available_subscriptions(self):
        """[summary]

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError

    def get_available_resource_groups(self, subscription_id: str):
        """[summary]

        Args:
            subscription_id (str): [description]

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError

    # -- Workspaces ----------------------------------------------------------

    def get_available_workspaces(self, subscription_id: str, resource_group: str):
        """[summary]

        Args:
            subscription_id (str): [description]
            resource_group (str): [description]

        Raises:
            NotImplementedError: [description]
        """

        raise NotImplementedError

    def get_available_workspace_names(self, subscription_id: str, resource_group: str):
        """[summary]

        Args:
            subscription_id (str): [description]
            resource_group (str): [description]

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError

    def get_workspace(self, project_meta: Dict[str, str]):
        """[summary]

        Args:
            subscription_id (str): [description]
            resource_group (str): [description]
            name (str): [description]

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError

    # -- Compute clusters ----------------------------------------------------

    def get_compute_clusters(self, workspace: Any):
        """[summary]

        Args:
            workspace (Any): [description]

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError

    def get_available_vm_sizes(self, workspace: Any) -> List[str]:
        """[summary]

        Args:
            workspace (Any): [description]

        Raises:
            NotImplementedError: [description]

        Returns:
            List[str]: [description]
        """
        raise NotImplementedError

    def create_compute_cluster(
        self,
        workspace: Any,
        name: str,
        vm_size: str,
        min_nodes: int,
        max_nodes: int,
        vnet_resource_group_name: str,
        vnet_name: str,
        subnet_name: str,
    ):
        """[summary]

        Args:
            workspace (Any): [description]
            name (str): [description]
            vm_size (str): [description]
            min_nodes (int): [description]
            max_nodes (int): [description]
            vnet_resource_group_name (str): [description]
            vnet_name (str): [description]
            subnet_name (str): [description]

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError

    # -- Contexts ------------------------------------------------------------

    def get_local_training_context(
        self, force_download: bool
    ) -> AbstractTrainingContext:
        """[summary]

        Args:
            force_download (bool): [description]

        Raises:
            NotImplementedError: [description]

        Returns:
            AbstractTrainingContext: [description]
        """
        raise NotImplementedError

    def get_cloud_training_context(self) -> AbstractTrainingContext:
        """[summary]

        Raises:
            NotImplementedError: [description]

        Returns:
            AbstractTrainingContext: [description]
        """
        raise NotImplementedError

    def submit_model(self, model: "Model", params: List[str]) -> SubmitContext:
        """[summary]

        Args:
            model (Model): [description]
            params (List[str]): [description]

        Raises:
            NotImplementedError: [description]

        Returns:
            SubmitContext: [description]
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import click
import click_spinner
from azureml.core import Experiment as AzExperiment
from azureml.core import Run as AzRun
from azureml.core import Workspace

from energinetml.backend import default_backend as backend
from energinetml.cli.utils import discover_model
from energinetml.core.model import Model, ModelArtifact

# -- CLI Command -------------------------------------------------------------
# TODO: Remove Azure-related things out in backend


@click.command()
@discover_model()
@click.option(
    "--run-id",
    "-r",
    "run_id",
    type=str,
    required=False,
    default=None,
    help="Run ID to release",
)
def release(run_id: str, model: Model) -> None:
    """Release a model."""

    project_meta = model.project.as_dict()

    workspace = backend.get_workspace(project_meta)

    if run_id is None:
        run_id = _get_run_id_if_omitted(workspace, model)

    # -- Download files ------------------------------------------------------

    click.echo("Releasing model using Run ID: %s" % run_id)

    with ModelArtifact.from_cloud(
        model.experiment, run_id, project_meta=project_meta
    ) as artifact:

        run = AzRun.get(workspace, run_id)
        # Can not use run.register_model() here, as its only able to copy
        # files from the Run's output folder, and we also want to package
        # the project files (from the restored snapshot).
        with click_spinner.spinner():
            model_tags = {k: v for k, v in run.tags.items() if not k.startswith("_")}
            model_tags.update({"run_id": run_id})

            model = backend.release_model(
                workspace=workspace,
                model_path=artifact.path,
                model_name=model.name,
                properties=run.properties,
                description="TODO Describe me!",
                run_id=run_id,
                tags=model_tags,
            )

            click.echo(f"Released new model {model.name} version {model.version}")


# -- Helper functions --------------------------------------------------------


def _get_run_id_if_omitted(workspace: Workspace, model: Model) -> str:
    """[summary]

    Args:
        workspace (Workspace): [description]
        model (Model): [description]

    Raises:
        click.Abort: [description]

    Returns:
        str: [description]
    """
    experiment = AzExperiment(workspace=workspace, name=model.experiment)

    latest_runs = list(AzRun.list(experiment=experiment))

    if not latest_runs:
        click.echo(f"No runs exists for experiment {model.experiment}")
        raise click.Abort()

    latest_run_id = latest_runs[0].id

    use_latest_run = click.confirm(
        (
            "You did not provide me with a Run ID. "
            f"Would you like to use the latest (Run ID: {latest_run_id})?"
        )
    )

    if use_latest_run is False:
        click.echo(
            "I do not know which run to release. Provide me with a "
            "Run ID using the -r/--run-id parameter."
        )
        raise click.Abort()

    return latest_run_id



from unittest.mock import ANY, Mock, PropertyMock, patch

from click.testing import CliRunner

from energinetml.cli.cluster.change import change
from tests.constants import WORKSPACE_NAME

# -- Tests -------------------------------------------------------------------


@patch("energinetml.cli.cluster.change.backend")
def test__change_cluster__no_clusters_exists__should_abort(backend_mock, model):
    """
    :param Mock backend_mock:
    :param Model model:
    """
    # Workaround: Can not mock property "name" of Mock object
    workspace = Mock()
    type(workspace).name = PropertyMock(return_value=WORKSPACE_NAME)

    backend_mock.get_workspace.return_value = workspace
    backend_mock.get_compute_clusters.return_value = []

    # Act
    result = CliRunner().invoke(cli=change, args=["--path", model.path])

    # Assert
    assert result.exit_code == 1


@patch("energinetml.cli.cluster.change.backend")
@patch("energinetml.cli.cluster.change.click.prompt")
def test__change_cluster__provide_a_cluster_that_does_not_exists__should_prompt_for_clister_and_update_model(  # noqa: E501
    prompt_mock, backend_mock, model_with_project
):
    """
    :param Mock prompt_mock:
    :param Mock backend_mock:
    :param Model model_with_project:
    """
    # Workaround: Can not mock property "name" of Mock object
    workspace = Mock()
    type(workspace).name = PropertyMock(return_value=WORKSPACE_NAME)

    cluster = Mock()
    cluster.vm_size = "cluster-vm-size"
    type(cluster).name = PropertyMock(return_value="an-existing-cluster")

    backend_mock.get_workspace.return_value = workspace
    backend_mock.get_compute_clusters.return_value = [cluster]

    prompt_mock.return_value = "an-existing-cluster"

    # Act
    result = CliRunner().invoke(
        cli=change,
        args=[
            "--path",
            model_with_project.path,
            "--cluster-name",
            "a-cluster-that-doesnt-exists",
        ],
    )

    # Assert
    assert result.exit_code == 0

    prompt_mock.assert_called_once_with(
        text="Please enter name of a compute cluster to use", type=ANY
    )



from dataclasses import dataclass
from unittest.mock import ANY, Mock, patch

import pytest
from azureml._common.exceptions import AzureMLException

from energinetml.azure.backend import AzureBackend
from energinetml.core.project import Project


@pytest.fixture
def azure_backend():
    azure_backend = AzureBackend()
    azure_backend._credential = Mock(return_value=Mock())
    yield azure_backend


class TestAzureBackend:

    # -- get_available_subscriptions() Tests -------------------------------------

    @patch("energinetml.azure.backend.SubscriptionClient")
    def test__get_available_subscriptions__should_return_subscriptions(
        self, subscription_client_mock, azure_backend
    ):
        """
        :param Mock subscription_client_mock:
        :param AzureBackend azure_backend:
        """
        subscriptions = [1, 2, 3]

        subscription_client_instance = Mock()
        subscription_client_instance.subscriptions.list.return_value = subscriptions
        subscription_client_mock.return_value = subscription_client_instance

        # Act
        return_value = azure_backend.get_available_subscriptions()

        # Assert
        assert return_value == subscriptions

    @patch("energinetml.azure.backend.SubscriptionClient")
    def test__get_available_subscriptions__raises_azure_ml_exception__should_raise_backend_exception(  # noqa: E501
        self, subscription_client_mock, azure_backend
    ):
        """
        :param Mock subscription_client_mock:
        :param AzureBackend azure_backend:
        """
        subscription_client_instance = Mock()
        subscription_client_instance.subscriptions.list.side_effect = AzureMLException(
            "x"
        )  # noqa: E501
        subscription_client_mock.return_value = subscription_client_instance

        # Act + Assert
        with pytest.raises(AzureBackend.BackendException):
            azure_backend.get_available_subscriptions()

    # -- get_available_resource_groups() Tests -----------------------------------

    @patch("energinetml.azure.backend.ResourceManagementClient")
    def test__get_available_resource_groups__should_return_resource_groups(
        self, subscription_client_mock, azure_backend
    ):
        """
        :param Mock subscription_client_mock:
        :param AzureBackend azure_backend:
        """
        resource_groups = [1, 2, 3]

        subscription_client_instance = Mock()
        subscription_client_instance.resource_groups.list.return_value = (
            resource_groups  # noqa: E501
        )
        subscription_client_mock.return_value = subscription_client_instance

        # Act
        return_value = azure_backend.get_available_resource_groups("subscription_id")

        # Assert
        assert return_value == resource_groups

    @patch("energinetml.azure.backend.ResourceManagementClient")
    def test__get_available_resource_groups__raises_azure_ml_exception__should_raise_backend_exception(  # noqa: E501
        self, subscription_client_mock, azure_backend
    ):
        """
        :param Mock subscription_client_mock:
        :param AzureBackend azure_backend:
        """
        subscription_client_instance = Mock()
        subscription_client_instance.resource_groups.list.side_effect = (
            AzureMLException("x")
        )
        subscription_client_mock.return_value = subscription_client_instance

        # Act + Assert
        with pytest.raises(AzureBackend.BackendException):
            azure_backend.get_available_resource_groups("subscription_id")

    # -- get_available_workspaces() Tests ----------------------------------------

    @patch("energinetml.azure.backend.Workspace")
    def test__get_available_workspaces__should_return_workspaces(
        self, workspace_mock, azure_backend
    ):
        """
        :param Mock workspace_mock:
        :param AzureBackend azure_backend:
        """
        workspaces_mapped = {
            "something1": ["workspace1", "workspace2"],
            "something2": ["workspace3", "workspace4"],
        }

        workspace_mock.list.return_value = workspaces_mapped

        # Act
        return_value = azure_backend.get_available_workspaces(
            "subscription_id", "resource_group"
        )

        # Assert
        assert return_value == ["workspace1", "workspace2", "workspace3", "workspace4"]

    @patch("energinetml.azure.backend.Workspace")
    def test__get_available_workspaces__raises_azure_ml_exception__should_raise_backend_exception(  # noqa: E501
        self, workspace_mock, azure_backend
    ):
        """
        :param Mock workspace_mock:
        :param AzureBackend azure_backend:
        """
        workspace_mock.list.side_effect = AzureMLException("x")

        # Act + Assert
        with pytest.raises(AzureBackend.BackendException):
            azure_backend.get_available_workspaces("subscription_id", "resource_group")

    # -- get_available_workspaces() Tests ----------------------------------------

    @patch.object(AzureBackend, "get_available_workspaces")
    def test__get_available_workspace_names__should_return_workspace_names(  # noqa: E501
        self, get_available_workspaces_mock, azure_backend
    ):
        """
        :param Mock get_available_workspaces_mock:
        :param AzureBackend azure_backend:
        """

        @dataclass
        class MockWorkspace:
            name: str

        get_available_workspaces_mock.return_value = [
            MockWorkspace("workspace1"),
            MockWorkspace("workspace2"),
            MockWorkspace("workspace3"),
            MockWorkspace("workspace4"),
        ]

        # Act
        return_value = azure_backend.get_available_workspace_names(
            "subscription_id", "resource_group"
        )

        # Assert
        assert return_value == ["workspace1", "workspace2", "workspace3", "workspace4"]

    # -- get_workspace() Tests ---------------------------------------------------

    @patch("energinetml.azure.backend.Workspace")
    def test__get_workspace__should_return_workspace(
        self, workspace_mock, azure_backend
    ):
        """
        :param Mock workspace_mock:
        :param AzureBackend azure_backend:
        """
        workspace = Mock()
        workspace_mock.get.return_value = workspace
        project_meta = {
            "subscription_id": "subscription_id",
            "resource_group": "resource_group",
            "workspace_name": "name",
        }

        # Act
        return_value = azure_backend.get_workspace(project_meta)

        # Assert
        workspace_mock.get.assert_called_once_with(
            auth=ANY,
            subscription_id="subscription_id",
            resource_group="resource_group",
            name="name",
        )

        assert return_value is workspace

    @patch("energinetml.azure.backend.Workspace")
    def test__get_workspace__should_raise_config_error(
        self, workspace_mock, azure_backend
    ):
        """
        :param Mock workspace_mock:
        :param AzureBackend azure_backend:
        """
        workspace = Mock()
        workspace_mock.get.return_value = workspace
        # Config missing key: 'subscription_id'
        project_meta = {"resource_group": "resource_group", "workspace_name": "name"}

        # Act + Assert
        with pytest.raises(Project.ConfigError):
            azure_backend.get_workspace(project_meta)

    @patch("energinetml.azure.backend.Workspace")
    def test__get_workspace__raises_azure_ml_exception__should_raise_backend_exception(  # noqa: E501
        self, workspace_mock, azure_backend
    ):
        """
        :param Mock workspace_mock:
        :param AzureBackend azure_backend:
        """
        workspace_mock.get.side_effect = AzureMLException("x")

        project_meta = {
            "subscription_id": "subscription_id",
            "resource_group": "resource_group",
            "workspace_name": "name",
        }
        # Act + Assert
        with pytest.raises(AzureBackend.BackendException):
            azure_backend.get_workspace(project_meta)

    # -- get_compute_clusters() Tests --------------------------------------------

    @patch("energinetml.azure.backend.AmlCompute")
    def test__get_compute_clusters__should_return_compute_clisters(
        self, aml_compute_mock, azure_backend
    ):
        """
        :param Mock aml_compute_mock:
        :param AzureBackend azure_backend:
        """
        compute_clusters = ["compute1", "compute2"]
        aml_compute_mock.list.return_value = compute_clusters

        # Act
        return_value = azure_backend.get_compute_clusters("workspace")

        # Assert
        aml_compute_mock.list.assert_called_once_with(workspace="workspace")

        assert return_value is compute_clusters

    @patch("energinetml.azure.backend.AmlCompute")
    def test__get_compute_clusters__raises_azure_ml_exception__should_raise_backend_exception(  # noqa: E501
        self, aml_compute_mock, azure_backend
    ):
        """
        :param Mock aml_compute_mock:
        :param AzureBackend azure_backend:
        """
        aml_compute_mock.list.side_effect = AzureMLException("x")

        # Act + Assert
        with pytest.raises(AzureBackend.BackendException):
            azure_backend.get_compute_clusters("workspace")

    # -- get_compute_clusters() Tests --------------------------------------------

    @patch("energinetml.azure.backend.AmlCompute")
    def test__get_available_vm_sizes__should_return_vm_sizes(
        self, aml_compute_mock, azure_backend
    ):
        """
        :param Mock aml_compute_mock:
        :param AzureBackend azure_backend:
        """
        vm_sizes = ["vm1", "vm2"]
        aml_compute_mock.supported_vmsizes.return_value = vm_sizes

        # Act
        return_value = azure_backend.get_available_vm_sizes("workspace")

        # Assert
        aml_compute_mock.supported_vmsizes.assert_called_once_with(
            workspace="workspace"
        )

        assert return_value is vm_sizes

    @patch("energinetml.azure.backend.AmlCompute")
    def test__get_available_vm_sizes__raises_azure_ml_exception__should_raise_backend_exception(  # noqa: E501
        self, aml_compute_mock, azure_backend
    ):
        """
        :param Mock aml_compute_mock:
        :param AzureBackend azure_backend:
        """
        aml_compute_mock.supported_vmsizes.side_effect = AzureMLException("x")

        # Act + Assert
        with pytest.raises(AzureBackend.BackendException):
            azure_backend.get_available_vm_sizes("workspace")



import os
import tempfile
from unittest.mock import Mock, patch

import pytest
from azureml import exceptions

from energinetml.azure.datasets import AzureMLDataStore, MLDataSet


class TestMLDataSet:
    def test__path(self):
        with tempfile.TemporaryDirectory() as path:
            uut = MLDataSet(name="NAME", mount_path=path)

            assert uut.path("some", "folder") == os.path.join(path, "some", "folder")

    @patch("energinetml.azure.datasets.open")
    def test__open(self, open_mock):
        """
        :param Mock open_mock:
        """
        file_object = Mock()
        open_mock.return_value = file_object

        with tempfile.TemporaryDirectory() as path:
            uut = MLDataSet(name="NAME", mount_path=path)

            return_value = uut.open(["some", "path"])

            assert return_value is file_object

    @patch("energinetml.azure.datasets.os.path.exists")
    def test__contains(self, exists_mock):
        """
        :param Mock exists_mock:
        """
        exists_mock.return_value = True

        with tempfile.TemporaryDirectory() as path:
            uut = MLDataSet(name="NAME", mount_path=path)

            return_value = uut.contains("some", "path")

            assert return_value is True


class TestAzureMLDataStore:
    @patch("energinetml.azure.datasets.AzureMLDataStore.load_azureml_dataset")
    @patch("energinetml.azure.datasets.AzureMLDataStore.mount")
    def test__from_model(self, mount_mock, load_azureml_dataset_mock):
        """
        :param Mock mount_mock:
        :param Mock load_azureml_dataset_mock:
        """
        workspace = Mock()
        force_download = True
        model = Mock()

        mount1 = Mock()
        mount2 = Mock()
        mount_mock.side_effect = (mount1, mount2)

        # Act
        instance = AzureMLDataStore.from_model(
            model=model,
            datasets=[("dataset1", "123"), ("dataset2", "321")],
            workspace=workspace,
            force_download=force_download,
        )

        # Assert
        assert isinstance(instance, AzureMLDataStore)

        assert load_azureml_dataset_mock.call_count == 2
        load_azureml_dataset_mock.assert_any_call(
            workspace=workspace, dataset_name="dataset1", dataset_version="123"
        )

        load_azureml_dataset_mock.assert_any_call(
            workspace=workspace, dataset_name="dataset2", dataset_version="321"
        )

        assert instance["dataset1"] is mount1
        assert instance["dataset2"] is mount2

    @patch("energinetml.azure.datasets.Dataset.get_by_name")
    @pytest.mark.parametrize(
        "dataset_version, actual_dataset_version", (("123", "123"), (None, "latest"))
    )
    def test__load_azureml_dataset__dataset_exists__should_return_dataset(
        self, get_by_name_mock, dataset_version, actual_dataset_version
    ):
        """
        :param Mock get_by_name_mock:
        :param str dataset_version:
        :param str actual_dataset_version:
        """
        dataset = Mock()
        get_by_name_mock.return_value = dataset
        workspace = Mock()

        # Act
        returned_value = AzureMLDataStore.load_azureml_dataset(
            workspace=workspace,
            dataset_name="datasetname",
            dataset_version=dataset_version,
        )

        # Assert
        assert returned_value is dataset

        get_by_name_mock.assert_called_once_with(
            workspace=workspace, name="datasetname", version=actual_dataset_version
        )

    @patch("energinetml.azure.datasets.Dataset.get_by_name")
    def test__load_azureml_dataset__azure_raise_user_error_exception__should_raise_data_set_not_found(  # noqa: E501
        self, get_by_name_mock
    ):
        """
        :param Mock get_by_name_mock:
        """
        get_by_name_mock.side_effect = exceptions._azureml_exception.UserErrorException(
            "x"
        )

        # Act + Assert
        with pytest.raises(AzureMLDataStore.DataSetNotFound):
            AzureMLDataStore.load_azureml_dataset(
                workspace=Mock(), dataset_name="datasetname", dataset_version="123"
            )

    def test__mount(self):
        with pytest.raises(NotImplementedError):
            AzureMLDataStore.mount(
                model=Mock(), azureml_dataset=Mock(), force_download=Mock()
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
from typing import Tuple, Union

import click

from energinetml.backend import default_backend as backend
from energinetml.cli.utils import discover_model
from energinetml.cli.utils.verify import verify_compute_cluster
from energinetml.core.backend import AbstractBackend
from energinetml.core.model import Model
from energinetml.settings import PACKAGE_NAME, PACKAGE_REQUIREMENT, PACKAGE_VERSION


@click.command()
@discover_model()
@click.argument("parameters", nargs=-1)
@click.option(
    "--wait/--nowait",
    default=True,
    help="Wait for experiment to complete before exiting (default: yes)",
)
@click.option(
    "--download", "-d", is_flag=True, help="Download output files after Run completed"
)
@click.option("--seed", "-s", required=False, default=None, type=str, help="Seed value")
def submit(
    parameters: Tuple[str, ...],
    wait: bool,
    download: bool,
    seed: Union[int, str],
    model: Model,
) -> None:
    """Submit a model to be trained in the cloud."""
    if download and not wait:
        click.echo("Can not use -d/--download together with --nowait")
        raise click.Abort()

    # -- Verify SDK requirement ----------------------------------------------

    # Verify requirements.txt contains energinet-ml-sdk
    if PACKAGE_NAME not in model.requirements:
        click.echo(
            (
                f"Could not find '{PACKAGE_NAME}' in the project's "
                f"requirements.txt file: {model.project.requirements_file_path}"
            )
        )
        click.echo("I suggest you add the following line:")
        click.echo("")
        click.echo(f"    {PACKAGE_REQUIREMENT}")
        click.echo("")
        raise click.Abort()

    required_sdk_version = model.requirements.get_version(PACKAGE_NAME)

    if required_sdk_version is None:
        click.echo(
            f"Using latest version of {PACKAGE_NAME} when training in the cloud."
        )
    else:
        if required_sdk_version < PACKAGE_VERSION:
            click.echo("=" * 79)
            click.echo(
                (
                    "WARNING: Your requirements.txt file contains a version of "
                    "%s (%s) which is older than your current installation (%s). "
                    "I suggest you update your requirements.txt to match "
                    "your current installation. "
                )
                % (PACKAGE_NAME, required_sdk_version, PACKAGE_VERSION)
            )
            click.echo("=" * 79)
        elif required_sdk_version > PACKAGE_VERSION:
            click.echo("=" * 79)
            click.echo(
                (
                    "WARNING: Your requirements.txt file contains a version of "
                    "%s (%s) which is newer than your current installation (%s). "
                    "I suggest you upgrade your installation using "
                    '"pip install --upgrade %s ...". '
                )
                % (PACKAGE_NAME, required_sdk_version, PACKAGE_VERSION, PACKAGE_NAME)
            )
            click.echo("=" * 79)

        click.echo(
            "Using {} version {} when training in the cloud.".format(
                PACKAGE_NAME, required_sdk_version
            )
        )

    # -- Verify computer cluster ---------------------------------------------
    verify_compute_cluster(backend, model, model.compute_target)
    verify_vm_size(backend, model)

    # -- Train Parameters ----------------------------------------------------

    if seed:
        parameters = list(parameters)
        parameters.extend(("--seed", seed))

    # -- Submit --------------------------------------------------------------

    context = backend.submit_model(model=model, params=parameters)

    # Stream output to console?
    if wait:
        click.echo("-" * 79)
        click.echo("Streaming Run log...")
        click.echo("-" * 79)

        try:
            context.wait_for_completion()
        except context.FailedToWait as e:
            click.echo("-" * 79)
            click.echo(str(e))
            click.echo("")
            raise click.Abort()

    # -- Post-Run ------------------------------------------------------------

    # Download output files?
    if download:
        click.echo("-" * 79)
        click.echo("Downloading files...")
        context.download_files()


def verify_vm_size(backend: AbstractBackend, model: Model):
    """Verifies that the VM size of the compute cluster matches
    the model's requirements.

    Args:
        backend (AbstractBackend): The backend to use.
        model (Model): The model that specifies the vm size

    Raises:
        click.Abort: If the compute cluster VM size does not match the
            model's requirements.
    """

    workspace = backend.get_workspace(model.project.as_dict())
    existing_clusters = backend.get_compute_clusters(workspace)
    existing_cluster_names = [c.name for c in existing_clusters]

    cluster_name = model.compute_target
    cluster = existing_clusters[existing_cluster_names.index(cluster_name)]

    if cluster.vm_size.casefold() != model.vm_size.casefold():
        click.echo(
            (
                f"WARNING: The VM size ({model.vm_size.casefold()}) "
                f"specified in {model._CONFIG_FILE_NAME}"
                " does not match the actual VM size of the "
                f"compute cluster ({cluster.vm_size.casefold()})."
            ),
            err=True,
        )
        raise click.Abort()



import os
import tempfile
from pathlib import Path

import pytest

from energinetml.core.files import FileMatcher


@pytest.fixture
def file_matcher():
    with tempfile.TemporaryDirectory() as path:
        yield FileMatcher(path=path)


class TestFileMatcher:
    def test__file_exists__should_return_correct_path(self):
        with tempfile.TemporaryDirectory() as path:
            Path(os.path.join(path, "f.json")).touch()
            Path(os.path.join(path, "f2.txt")).touch()
            os.makedirs(os.path.join(path, "subfolder"))
            Path(os.path.join(path, "subfolder", "f3.json")).touch()
            Path(os.path.join(path, "subfolder", "f4.txt")).touch()

            uut = FileMatcher(
                root_path=path,
                include=["**/*.json", "**/*.txt"],
                exclude=["subfolder/*.txt"],
            )

            matches = list(uut)

            assert len(matches) == 3
            assert "f.json" in matches
            assert "f2.txt" in matches
            assert os.path.join("subfolder", "f3.json") in matches



import tempfile
from unittest.mock import PropertyMock, patch

from click.testing import CliRunner

from energinetml.cli.model.init import init_model
from energinetml.core.model import Model
from energinetml.core.project import MachineLearningProject

# Project
PROJECT_NAME = "PROJECTNAME"
SUBSCRIPTION_ID = "SUBSCRIPTION-ID"
RESOURCE_GROUP = "RESOURCE-GROUP"
WORKSPACE_NAME = "WORKSPACE-NAME"
VNET = "VNET"
SUBNET = "SUBNET"


# Model
MODEL_NAME = "MODELNAME"
EXPERIMENT = "EXPERIMENT"
COMPUTE_TARGET = "COMPUTE-TARGET"
VM_SIZE = "VM-SIZE"
DATASETS = ["iris", "hades:2"]
FEATURES = ["feature1", "feature2"]
PARAMETERS = {"param1": "value1", "param2": "value2"}


# -- Tests -------------------------------------------------------------------


@patch("energinetml.cli.model.init.os.path.isfile")
@patch("energinetml.cli.model.init.click.confirm")
@patch.object(Model, "create", new_callable=PropertyMock)
def test__init_model__deny_override_existing_files__should_abort(
    model_create_mock, confirm_mock, isfile_mock
):
    """
    :param Mock model_create_mock:
    :param Mock confirm_mock:
    :param Mock isfile_mock:
    """
    isfile_mock.return_value = True
    confirm_mock.return_value = False

    runner = CliRunner()

    with tempfile.TemporaryDirectory() as path:
        MachineLearningProject.create(
            path=path,
            name=PROJECT_NAME,
            subscription_id=SUBSCRIPTION_ID,
            resource_group=RESOURCE_GROUP,
            workspace_name=WORKSPACE_NAME,
            vnet_name=VNET,
            subnet_name=SUBNET,
        )

        # Act
        result = runner.invoke(
            cli=init_model, args=["--path", path, "--name", "mymodel"]
        )

        # Assert
        assert result.exit_code == 1


@patch("energinetml.cli.model.init.click.confirm")
@patch("energinetml.cli.model.init.create_cluster")
def test__init_model__no_name_provided__should_prompt_for_name_and_create_model_and_new_cluster(  # noqa: E501
    create_cluster_mock, confirm_mock
):
    """
    :param Mock create_cluster_mock:
    :param Mock confirm_mock:
    """
    confirm_mock.return_value = True

    runner = CliRunner()

    with tempfile.TemporaryDirectory() as path:
        MachineLearningProject.create(
            path=path,
            name=PROJECT_NAME,
            subscription_id=SUBSCRIPTION_ID,
            resource_group=RESOURCE_GROUP,
            workspace_name=WORKSPACE_NAME,
            vnet_name=VNET,
            subnet_name=SUBNET,
        )

        # Act
        result = runner.invoke(
            cli=init_model, args=["--path", path], input="MODEL-NAME\nnew\n"
        )

        # Assert
        assert result.exit_code == 0

        create_cluster_mock.assert_called_once()


@patch("energinetml.cli.model.init.click.confirm")
@patch("energinetml.cli.model.init.change_cluster")
def test__init_model__no_name_provided__should_prompt_for_name_and_create_model_and_change_cluster(  # noqa: E501
    change_cluster_mock, confirm_mock
):
    """
    :param Mock change_cluster_mock:
    :param Mock confirm_mock:
    """
    confirm_mock.return_value = True

    runner = CliRunner()

    with tempfile.TemporaryDirectory() as path:
        MachineLearningProject.create(
            path=path,
            name=PROJECT_NAME,
            subscription_id=SUBSCRIPTION_ID,
            resource_group=RESOURCE_GROUP,
            workspace_name=WORKSPACE_NAME,
            vnet_name=VNET,
            subnet_name=SUBNET,
        )

        # Act
        result = runner.invoke(
            cli=init_model, args=["--path", path], input="MODEL-NAME\nexisting\n"
        )

        # Assert
        assert result.exit_code == 0

        change_cluster_mock.assert_called_once()



import os
import tempfile

import pytest

from energinetml.core.configurable import Configurable, locate_file_upwards

# This satisfies SonarCloud
SOME_FILE_NAME = "somefile.txt"


@pytest.fixture
def configurable():
    with tempfile.TemporaryDirectory() as path:
        yield Configurable(path=path)


# -- Tests -------------------------------------------------------------------


def test__locate_file_upwards__file_exists__should_return_correct_path():
    with tempfile.TemporaryDirectory() as path:
        starting_point = os.path.join(path, "some", "subfolder")
        target_file = os.path.join(path, SOME_FILE_NAME)

        os.makedirs(starting_point)

        with open(target_file, "w") as f:
            f.write("hello!")

        assert locate_file_upwards(starting_point, SOME_FILE_NAME) == target_file


def test__locate_file_upwards__file_does_not_exists__should_return_none():
    with tempfile.TemporaryDirectory() as path:
        starting_point = os.path.join(path, "some", "subfolder")

        os.makedirs(starting_point)

        assert locate_file_upwards(starting_point, SOME_FILE_NAME) is None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

import os
import re
import subprocess
import sys
import tempfile
from typing import Dict

import click
from jinja2 import Template

from energinetml.settings import (
    DEFAULT_ENCODING,
    TEMPLATES_GIT_URL,
    TEMPLATES_IP_WHITELIST,
    TEMPLATES_SUBNET_WHITELIST,
)

FILES_TO_COPY = (".gitignore",)
FOLDERS_TO_COPY = (".azuredevops", "terraform")


# -- Input parsing and validation --------------------------------------------


def _parse_input_path(ctx: click.Context, param: click.Parameter, value: str) -> str:
    """[summary]

    Args:
        ctx (click.Context): [description]
        param (click.Parameter): [description]
        value (str): [description]

    Raises:
        click.Abort: [description]

    Returns:
        str: [description]
    """
    if value is None:
        value = os.path.abspath(
            click.prompt(
                text="Enter project location",
                default=os.path.abspath("."),
                type=click.Path(dir_okay=True, resolve_path=True),
            )
        )

    # Path points to a file?
    if os.path.isfile(value):
        click.echo("Failed to init project infrastructure.")
        click.echo(
            "The path you provided me with points to a file, and not a "
            "folder. I need a folder to put the project files in. "
            "Check your -p/--path parameter."
        )
        click.echo("You provided me with: %s" % value)
        raise click.Abort()

    return value


def _parse_input_project_name(
    ctx: click.Context, param: click.Parameter, value: str
) -> str:
    """[summary]

    Args:
        ctx (click.Context): [description]
        param (click.Parameter): [description]
        value (str): [description]

    Returns:
        str: [description]
    """
    while not value or not re.findall(r"^[a-z][a-z0-9]{,10}$", value):
        if value is not None:
            click.echo("Invalid name provided")

        click.echo(
            "Provisioning cloud resources requires a name for your project "
            "which contains 11 (or less) characters. This name is used as part "
            "of the resource names, can only contain lower case letters "
            "and numbers, and must start with a letter."
        )

        value = click.prompt(text="Please enter a project name", type=click.STRING)

    return value


def _parse_input_resource_group(
    ctx: click.Context, param: click.Parameter, value: str
) -> str:
    """[summary]

    Args:
        ctx (click.Context): [description]
        param (click.Parameter): [description]
        value (str): [description]

    Returns:
        str: [description]
    """
    if value is None:
        value = click.prompt(text="Enter resource group name", type=str)

    return value


def _parse_input_service_connection(
    ctx: click.Context, param: click.Parameter, value: str
) -> str:
    """[summary]

    Args:
        ctx (click.Context): [description]
        param (click.Parameter): [description]
        value (str): [description]

    Returns:
        str: [description]
    """
    if value is None:
        value = click.prompt(
            text="Enter Azure DevOps service connection name", type=str
        )

    return value


# -- CLI Command -------------------------------------------------------------


@click.command()
@click.option(
    "--path",
    "-p",
    default=None,
    type=click.Path(dir_okay=True, resolve_path=True),
    callback=_parse_input_path,
    help="Project path (default to current)",
)
@click.option(
    "--name",
    "-n",
    "project_name",
    default=None,
    type=str,
    callback=_parse_input_project_name,
    help="Project name",
)
@click.option(
    "--resource-group",
    "-g",
    "resource_group",
    default=None,
    type=str,
    callback=_parse_input_resource_group,
    help="Azure resource group",
)
@click.option(
    "--service-connection",
    "-s",
    "service_connection",
    default=None,
    type=str,
    callback=_parse_input_service_connection,
    help="Azure DevOps service connection name",
)
def init_infrastructure(
    path: str, project_name: str, resource_group: str, service_connection: str
):
    """Create infrastructure files for a new AnalyticsOps project."""

    # -- Clone repo ----------------------------------------------------------

    click.echo("-" * 79)
    click.echo(
        "NOTICE: We need to clone a Git repository containing the "
        "necessary template files. This requires Git to be "
        "installed on your system."
    )
    click.echo("-" * 79)

    # -- Clone repo ----------------------------------------------------------

    with tempfile.TemporaryDirectory() as clone_path:
        _clone_pipeline_files(clone_path=clone_path)

        _parse_cloned_files(
            clone_path=clone_path,
            target_path=path,
            project_name=project_name,
            resource_group=resource_group,
            service_connection=service_connection,
        )


def _clone_pipeline_files(clone_path: str) -> None:
    """[summary]

    Args:
        clone_path (str): [description]

    Raises:
        click.Abort: [description]
    """
    click.echo(f'Cloning "{TEMPLATES_GIT_URL}" into "{clone_path}"')

    try:
        subprocess.check_call(
            args=["git", "clone", TEMPLATES_GIT_URL, clone_path],
            stdout=sys.stdout,
            stderr=subprocess.STDOUT,
        )
    except subprocess.CalledProcessError:
        raise click.Abort()


def _parse_cloned_files(
    clone_path: str,
    target_path: str,
    project_name: str,
    resource_group: str,
    service_connection: str,
) -> None:
    """[summary]

    Args:
        clone_path (str): [description]
        target_path (str): [description]
        project_name (str): [description]
        resource_group (str): [description]
        service_connection (str): [description]
    """
    env = {
        "projectName": project_name,
        "resourceGroup": resource_group,
        "serviceConnection": service_connection,
        "subnetWhitelist": TEMPLATES_SUBNET_WHITELIST,
        "ipWhitelist": TEMPLATES_IP_WHITELIST,
    }

    files_to_copy = []
    files_to_copy.extend(os.path.join(clone_path, f) for f in FILES_TO_COPY)

    for folder in FOLDERS_TO_COPY:
        for root, subdirs, files in os.walk(os.path.join(clone_path, folder)):
            for file_name in files:
                files_to_copy.append(os.path.join(root, file_name))

    for file_abs_path in files_to_copy:
        rel_path = os.path.relpath(file_abs_path, clone_path)
        dst_path = os.path.join(target_path, rel_path)

        _copy_file(src=file_abs_path, dst=dst_path, env=env)


def _copy_file(src: str, dst: str, env: Dict[str, str]) -> None:
    """[summary]

    Args:
        src (str): [description]
        dst (str): [description]
        env (Dict[str, str]): [description]
    """
    dst_folder = os.path.split(dst)[0]

    with open(src, encoding=DEFAULT_ENCODING) as f:
        template = Template(f.read())
        rendered = template.render(**env)

    if not os.path.isdir(dst_folder):
        os.makedirs(dst_folder)

    with open(dst, "w", encoding=DEFAULT_ENCODING) as f:
        f.write(rendered)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import os
import subprocess
import sys

import click

from energinetml.cli.utils import (
    parse_input_path,
    parse_input_project_name,
    parse_input_resource_group,
    parse_input_service_connection,
    parse_input_webapp_kind,
    parse_pipeline_repo_version,
)
from energinetml.core.project import WebAppProject, WebAppProjectKind

PROJECT_FILES = ("project.json", "requirements.txt")


# -- CLI Command -------------------------------------------------------------


@click.command()
@click.option(
    "--path",
    "-p",
    default=None,
    type=click.Path(dir_okay=True, resolve_path=True),
    callback=parse_input_path(PROJECT_FILES),
    help="Project path (default to current)",
)
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    type=str,
    callback=parse_input_project_name(),
    help="Project name",
)
@click.option(
    "--resource-group",
    "-r",
    "resource_group",
    required=False,
    default=None,
    type=str,
    callback=parse_input_resource_group(),
    help="Azure Resource Group",
)
@click.option(
    "--service-connection",
    "-s",
    "service_connection",
    required=False,
    default=None,
    type=str,
    callback=parse_input_service_connection(),
    help="Azure DevOps Service Connection name",
)
@click.option(
    "--pipelines/--no-pipelines",
    default=None,
    help="Whether or not to setup DevOps pipelines",
)
@click.option(
    "--kind",
    "-k",
    "kind",
    required=False,
    default=None,
    type=click.Choice(["ASGI", "WSGI"]),
    callback=parse_input_webapp_kind(),
    help=(
        "Kind of web server to use (ASGI or WSGI). "
        "ASGI for flask, Django, or Falcon. "
        "WSGI for fastapi."
    ),
)
@click.option(
    "--pipeline-repo-version",
    "-pv",
    "pipeline_repo_version",
    required=False,
    default=None,
    type=str,
    callback=parse_pipeline_repo_version(),
    help="Version of the public Pipeline repository",
)
def init(
    path: str,
    name: str,
    resource_group: str,
    service_connection: str,
    pipeline_repo_version: str,
    kind: WebAppProjectKind,
    pipelines: bool,
) -> None:
    """Create a new, empty web app project."""

    # -- Create project ------------------------------------------------------

    project = WebAppProject.create(path=path, name=name, kind=kind)

    click.echo("Initialized the project at: %s" % path)
    click.echo("-" * 79)

    # -- Clone repo ----------------------------------------------------------

    click.echo("-" * 79)
    click.echo(
        "NOTICE: We need to clone a Git repository containing the "
        "necessary template files. This requires Git to be "
        "installed on your system."
    )
    click.echo("-" * 79)

    project.get_template_resolver().resolve(
        project_root_path=path,
        project_name=name,
        service_connection=service_connection or "ENTER SERVICE CONNECTION NAME",
        resource_group=resource_group or "ENTER RESOURCE GROUP",
        pipeline_repo_version=pipeline_repo_version
        or "ENTER PIPELINE REPOSITORY VERSION",
    )

    click.echo("-" * 79)

    # -- Create DevOps pipelines ---------------------------------------------

    click.echo("Creating Azure DevOps pipelines")

    if pipelines:
        repo = _get_git_repository_name(project)

        _create_pipeline(
            name=f"{project.name} infrastructure",
            yaml_path=".azuredevops/infrastructure.yml",
            repo=repo,
        )

        _create_pipeline(
            name=f"{project.name} deploy",
            yaml_path=".azuredevops/deploy.yml",
            repo=repo,
        )


def _get_git_repository_name(project: WebAppProject) -> str:
    """[summary]

    Args:
        project (WebAppProject): [description]

    Returns:
        str: [description]
    """
    path = os.getcwd()
    os.chdir(project.path)
    command = ("git", "config", "--get", "remote.origin.url")
    output = subprocess.check_output(command, universal_newlines=True)
    os.chdir(path)
    return output.rsplit("/")[-1].strip()


def _create_pipeline(name: str, yaml_path: str, repo: str):
    """[summary]

    Args:
        name (str): [description]
        yaml_path (str): [description]
        repo (str): [description]
    """
    click.echo(f"Setting up DevOps pipeline: {name}")

    command = [
        "az",
        "pipelines",
        "create",
        "--name",
        f"{name}",
        "--branch",
        "master",
        "--org",
        "https://dev.azure.com/energinet/",
        "--project",
        "AnalyticsOps",
        "--repository",
        repo,
        "--yaml-path",
        yaml_path,
        "--repository-type",
        "tfsgit",
        "--skip-first-run",
        "true",
    ]

    subprocess.check_call(command, stdout=sys.stdout, stderr=subprocess.STDOUT)



import tempfile
from random import random
from unittest.mock import Mock

import pytest

from energinetml.core.model import Model, TrainedModel, import_model_class
from energinetml.core.project import MachineLearningProject
from tests.constants import (
    COMPUTE_TARGET,
    DATASETS,
    DATASETS_CLOUD,
    DATASETS_LOCAL,
    EXPERIMENT,
    FEATURES,
    FILES_INCLUDE,
    MODEL_NAME,
    PARAMETERS,
    PROJECT_NAME,
    RESOURCE_GROUP,
    SUBNET,
    SUBSCRIPTION_ID,
    VM_SIZE,
    VNET,
    WORKSPACE_NAME,
)


def create_model(path: str, files_include=FILES_INCLUDE):
    return Model.create(
        path=f"{path}/user_model_{int(random()*100000)}",
        name=MODEL_NAME,
        experiment=EXPERIMENT,
        compute_target=COMPUTE_TARGET,
        vm_size=VM_SIZE,
        datasets=DATASETS,
        datasets_local=DATASETS_LOCAL,
        datasets_cloud=DATASETS_CLOUD,
        features=FEATURES,
        parameters=PARAMETERS,
        files_include=files_include,
    )


def create_project(path: str):
    return MachineLearningProject.create(
        path=path + "/user_model_test",
        name=PROJECT_NAME,
        subscription_id=SUBSCRIPTION_ID,
        resource_group=RESOURCE_GROUP,
        workspace_name=WORKSPACE_NAME,
        vnet_name=VNET,
        subnet_name=SUBNET,
    )


@pytest.fixture
def model_path():
    with tempfile.TemporaryDirectory() as path:
        project = create_project(path)

        model = create_model(path=project.default_model_path(MODEL_NAME))

        trained_model = TrainedModel(
            model="123", params={"asd": 123}, features=FEATURES
        )
        Model.dump(model.trained_model_path, trained_model)

        yield model.path


@pytest.fixture
def user_model():
    with tempfile.TemporaryDirectory() as path:
        model = create_model(path)
        model_class = import_model_class(model.path, model.parent_levels)
        yield model_class.from_directory(model.path)


@pytest.fixture
def model():
    with tempfile.TemporaryDirectory() as path:
        yield create_model(path)


@pytest.fixture
def model_with_project():
    with tempfile.TemporaryDirectory() as path:
        project = create_project(path)
        yield create_model(project.default_model_path(MODEL_NAME))


@pytest.fixture
def clusters_mock():
    cluster1 = Mock()
    cluster1.name = COMPUTE_TARGET
    cluster1.vm_size = VM_SIZE
    return [cluster1]


# -- Smoketest command-line options ------------------------------------------


def pytest_addoption(parser):
    """
    Adds command-line options to "pytest" command which will
    become available when running tests. Used for smoke testing.
    """
    parser.addoption("--path", action="store", default=None)
    parser.addoption("--subscription-id", action="store", default=None)
    parser.addoption("--subscription-name", action="store", default=None)
    parser.addoption("--resource-group", action="store", default=None)
    parser.addoption("--service-connection", action="store", default=None)
    parser.addoption("--workspace-name", action="store", default=None)
    parser.addoption("--project-name", action="store", default=None)
    parser.addoption("--model-name", action="store", default=None)
    parser.addoption("--deployment-base-url", action="store", default=None)
    parser.addoption("--sdk-version", action="store", default=None)


@pytest.fixture
def path(pytestconfig):
    """
    Used by ML smoketests.
    """
    return pytestconfig.getoption("--path")


@pytest.fixture
def subscription_id(pytestconfig):
    """
    Used by ML smoketests.
    """
    return pytestconfig.getoption("--subscription-id")


@pytest.fixture
def subscription_name(pytestconfig):
    """
    Used by ML smoketests.
    """
    return pytestconfig.getoption("--subscription-name")


@pytest.fixture
def resource_group(pytestconfig):
    """
    Used by ML AND Web-App smoketests.
    """
    return pytestconfig.getoption("--resource-group")


@pytest.fixture
def service_connection(pytestconfig):
    """
    Used by Web-App smoketests.
    """
    return pytestconfig.getoption("--service-connection")


@pytest.fixture
def workspace_name(pytestconfig):
    """
    Used by ML smoketests.
    """
    return pytestconfig.getoption("--workspace-name")


@pytest.fixture
def project_name(pytestconfig):
    """
    Used by ML AND Web-App smoketests.
    """
    return pytestconfig.getoption("--project-name")


@pytest.fixture
def model_name(pytestconfig):
    """
    Used by ML smoketests.
    """
    return pytestconfig.getoption("--model-name")


@pytest.fixture
def deployment_base_url(pytestconfig):
    """
    Used by ML smoketests.
    """
    return pytestconfig.getoption("--deployment-base-url")


@pytest.fixture
def sdk_version(pytestconfig):
    """
    Used by ML smoketests.
    """
    return pytestconfig.getoption("--sdk-version")


@pytest.fixture
def prediction_input():
    """
    Used by ML smoketests.
    """
    return {"inputs": [{"features": {"age": 20}}, {"features": {"age": 40}}]}


@pytest.fixture
def prediction_output():
    """
    Used by ML smoketests.
    """
    return ["no", "yes"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This is the NewModel class which defines the training logic and
inference logic.
"""

import typing

from energinetml import MetricsLogger, Model, PredictionInput, TrainedModel
from energinetml.azure.datasets import MLDataStore


class NewModel(Model):
    """EML SDK model class. This class captures training logic and inference logic."""

    def train(
        self,
        datasets: MLDataStore,
        logger: MetricsLogger,
        seed: int,
        **params: typing.Dict[str, typing.Any]
    ) -> TrainedModel:
        """Define your training logic.

        Args:
            datasets (MLDataStore): A reference to the data set.
            logger (MetricsLogger): The logger argument handle metric logging, etc.
            seed (typing.Any): The seed argument tries to obtain a deterministic
            environment for model experiments.

        Returns:
            TrainedModel: A serializable object of the train model.
        """

        # TODO Train your model here and return it

        raise NotImplementedError

        return TrainedModel()

    def predict(
        self, trained_model: TrainedModel, input_data: PredictionInput, identifier: str
    ) -> typing.List[typing.Any]:
        """Define your prediction logic.

        Args:
            trained_model (TrainedModel): Your trained model object.
            input_data (PredictionInput): Data used for inference.
            identifier (str): A unique identifier which refers to a specific model in
            the model obejct.

        Returns:
            typing.List[typing.Any]: The output(s) of the model.
        """

        # TODO Predict using the trained model and return the prediction

        raise NotImplementedError



import pytest

from energinetml.core.submitting import SubmitContext


@pytest.fixture
def submit_context():
    yield SubmitContext()


class TestSubmitContext:
    def test__submit_model(self, submit_context):
        """
        :param SubmitContext submit_context:
        """
        with pytest.raises(NotImplementedError):
            submit_context.submit_model()

    def test__wait_for_completion(self, submit_context):
        """
        :param SubmitContext submit_context:
        """
        with pytest.raises(NotImplementedError):
            submit_context.wait_for_completion()

    def test__download_files(self, submit_context):
        """
        :param SubmitContext submit_context:
        """
        with pytest.raises(NotImplementedError):
            submit_context.download_files()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]"""

import importlib
import inspect
import json
import os
import pickle
import random
import shutil
import sys
import tempfile
import zipfile
from dataclasses import dataclass, field
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    ContextManager,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

import pandas as pd
import pydantic

from energinetml.azure.datasets import MLDataStore
from energinetml.backend import default_backend as backend
from energinetml.core.configurable import Configurable, locate_file_upwards
from energinetml.core.files import FileMatcher, temporary_folder
from energinetml.core.logger import MetricsLogger
from energinetml.core.predicting import PredictionController, PredictResponse
from energinetml.core.project import MachineLearningProject, Project
from energinetml.core.requirements import RequirementList
from energinetml.core.validation import Validator
from energinetml.settings import (
    DEFAULT_RELATIVE_ARTIFACT_PATH,
    EMPTY_MODEL_TEMPLATE_DIR,
)

if TYPE_CHECKING:
    from energinetml.core.predicting import PredictionInput

# Constants
# TODO Move to settings.py?
REQUIREMENTS_FILE_NAME = "requirements.txt"
DEFAULT_FILES_INCLUDE = ["**/*.py", "model.json", REQUIREMENTS_FILE_NAME]
DEFAULT_FILES_EXCLUDE = []


@dataclass
class Model(Configurable):
    """
    Class for holding a model.

    Attributes:
        name: Name of the model
        experiment: Name of the experiment
        compute_target: Compute target
        vm_size: Size of the virtual machine targeted
    """

    name: str
    experiment: str
    compute_target: str
    vm_size: str
    datasets: List[str] = field(default_factory=list)
    datasets_local: List[str] = field(default_factory=list)
    datasets_cloud: List[str] = field(default_factory=list)
    parameters: Dict[str, Any] = field(default_factory=dict)
    parameters_local: Dict[str, Any] = field(default_factory=dict)
    parameters_cloud: Dict[str, Any] = field(default_factory=dict)
    features: List[str] = field(default_factory=list)
    files_include: List[str] = field(default_factory=list)
    files_exclude: List[str] = field(default_factory=list)

    # Constants
    _CONFIG_FILE_NAME = "model.json"
    _SCRIPT_FILE_NAME = "model.py"
    _TRAINED_MODEL_FILE_NAME = "model.pkl"
    _META_FILE_NAME = "meta.json"
    _REQUIREMENTS_FILE_NAME = REQUIREMENTS_FILE_NAME

    @classmethod
    def create(cls, *args: Any, **kwargs: Dict[str, Any]) -> "Model":
        """[summary]

        Returns:
            Model: [description]
        """
        model = super().create(*args, **kwargs)

        # Copy template files
        for filename in os.listdir(EMPTY_MODEL_TEMPLATE_DIR):
            src = os.path.join(EMPTY_MODEL_TEMPLATE_DIR, filename)
            dst = os.path.join(model.path, filename)
            if os.path.isfile(src):
                shutil.copyfile(src, dst)

        return model

    @cached_property
    def project(self) -> Union[MachineLearningProject, None]:
        """Returns the Project which this model belongs to.

        Returns:
            Union[MachineLearningProject, None]
        """

        try:
            return MachineLearningProject.from_directory(self.path)
        except MachineLearningProject.ConfigNotFound:
            return None

    @cached_property
    def datasets_parsed(self) -> "ModelDatasets":
        """[summary]

        Returns:
            ModelDatasets
        """
        return ModelDatasets(self)

    @property
    def trained_model_path(self) -> str:
        """Returns path to the trained model."""
        return self.get_file_path(
            DEFAULT_RELATIVE_ARTIFACT_PATH, self._TRAINED_MODEL_FILE_NAME
        )

    @property
    def parent_levels(self) -> int:
        """
        Returns the number of parent levels.

        This is relevant in regards to support of relative imports.

        Returns:
            int
        """

        counts = [dir.count("..") for dir in self.files_include]
        relative_import_levels = max(counts)

        return relative_import_levels + 1

    @property
    def artifact_path(self) -> str:
        """Path to the artifact folder of a trained model. The folder should contain
        the model.pkl file.

        Returns:
            str
        """
        return self.get_file_path(DEFAULT_RELATIVE_ARTIFACT_PATH)

    @property
    def data_folder_path(self) -> str:
        """Returns path to the data folder.

        Returns:
            str
        """
        return self.get_file_path("data")

    @property
    def requirements_file_path(self) -> str:
        """Absolute path to requirements.txt file.

        Returns:
            str
        """
        return self.get_file_path(self._REQUIREMENTS_FILE_NAME)

    @cached_property
    def requirements(self) -> RequirementList:
        """[summary]

        Returns:
            RequirementList
        """
        if os.path.isfile(self.requirements_file_path):
            return RequirementList.from_file(self.requirements_file_path)
        elif self.project:
            return self.project.requirements
        else:
            return RequirementList()

    @property
    def files(self) -> FileMatcher:
        """Returns an iterable of files to include when submitting a model to
        the cloud. These are the files necessary to run training in the cloud.

        File paths are relative to model root.

        NB: Does NOT include requirements.txt !!!

        Returns:
            FileMatcher
        """
        return FileMatcher(
            root_path=self.path,
            include=self.files_include,
            exclude=self.files_exclude,
            recursive=True,
        )

    @property
    def module_name(self) -> str:
        """The name of the module to which the model belongs.

        The last part of the name is excluded, since it is always `model`
        and is unneeded.

        Returns:
            str
        """
        module_parts = self.__module__.split(".")[:-1]

        return ".".join(module_parts)

    def temporary_folder(
        self, include_trained_model: bool = False
    ) -> ContextManager[None]:
        """Returns a context manager which creates a temporary folder on the
        filesystem and copies all model's files into the folder including
        the project's requirements.txt file (if it exists).

        Example:
            >>> with model.temporary_folder() as temp_path:
            >>>    # files are available in temp_path

        Args:
            include_trained_model (bool, optional): [description]. Defaults to False.

        Returns:
                ContextManager[str]
        """
        files_to_copy = []

        # Model-specific files (relative to model root)
        path = os.path.normpath(self.path)
        relevant_parent_folders = path.split(os.sep)[-self.parent_levels :]
        parent_dir = os.path.join(*relevant_parent_folders)

        for relative_path in self.files:
            files_to_copy.append(
                (
                    self.get_file_path(relative_path),
                    os.path.join(parent_dir, relative_path),
                )
            )

        # Trained model (model.pkl) if necessary
        if include_trained_model:
            files_to_copy.append(
                (
                    self.trained_model_path,
                    self.get_relative_file_path(self.trained_model_path),
                )
            )

        # requirements.txt from this folder or project folder
        if os.path.isfile(self.requirements_file_path):
            files_to_copy.append(
                (self.requirements_file_path, self._REQUIREMENTS_FILE_NAME)
            )
        elif self.project and os.path.isfile(self.project.requirements_file_path):
            files_to_copy.append(
                (self.project.requirements_file_path, self._REQUIREMENTS_FILE_NAME)
            )

        return temporary_folder(files_to_copy)

    # -- Partially abstract interface ----------------------------------------

    # The following methods are meant to be overwritten by inherited classes
    # if necessary. Some can be omitted, and will return default values.

    def extra_tags(self) -> Dict[str, Any]:
        """TODO: What is this function for?

        Returns:
            Dict[str, Any]: [description]
        """
        return {}

    def generate_seed(self) -> int:
        """Generates a random seed between :math:`0` and :math:`10^9`.

        Returns:
            int: A random number between :math:`0` and :math:`10^9`.
        """
        return random.randint(0, 10**9)

    def train(
        self,
        datasets: MLDataStore,
        logger: MetricsLogger,
        seed: int,
        **params: Dict[str, Any],
    ) -> None:
        """Define your training logic.

        Args:
            datasets (MLDataStore): A reference to the available datasets.
            logger (MetricsLogger): The logger argument handle metric logging, etc.
            seed (typing.Any): The seed argument tries to obtain a deterministic
                environment for model experiments.
        """
        raise NotImplementedError

    def predict(
        self,
        trained_model: "TrainedModel",
        input_data: "PredictionInput",
        identifier: str,
    ) -> None:
        """Define your prediction logic.

        Args:
            trained_model (TrainedModel): Your trained model object.
            input_data (PredictionInput): Data used for inference.
            identifier (str): A unique identifier which refers to a specific model in
                the model object.
        """
        raise NotImplementedError

    @staticmethod
    def load(fp: str) -> "TrainedModel":
        """Load trained model from filepath.

        Args:
            fp (str): [description]

        Raises:
            TrainedModel.Invalid: [description]

        Returns:
            TrainedModel: [description]
        """
        with open(fp, "rb") as f:
            try:
                loaded_model = pickle.load(f)

                if not isinstance(loaded_model, TrainedModel):
                    raise TrainedModel.Invalid(
                        f"The file at {fp}"
                        f"does not contain a {TrainedModel.__name__} object."
                    )

                loaded_model.verify()

                return loaded_model
            except pickle.UnpicklingError:
                raise TrainedModel.Invalid("Failed to load trained model.")

    @staticmethod
    def dump(file_path: str, trained_model: "TrainedModel") -> None:
        """Dump :ref:`TrainedModel` to pickle-file in file path.

        Args:
            file_path (str): The filepath to dump the pickle-file to.
            trained_model (TrainedModel): The trained model to dump.
        """
        folder = os.path.split(file_path)[0]
        if not os.path.isdir(folder):
            os.makedirs(folder)
        with open(file_path, "wb") as f:
            pickle.dump(trained_model, f)


class ModelDatasets:
    """
    A wrapper for parsing datasets in string-format to specific
    name and version. Distinguishes between datasets for local
    and cloud training.
    """

    def __init__(self, model: Model):
        """[summary]

        Args:
            model (Model): [description]
        """
        self.model = model

    def _parse_datasets(self, datasets: List[str]) -> Iterable[Tuple[str, str]]:
        """Parses datasets in the format of either "name" or "name:version"
        and returns an iterable of (name, version), where version is
        optional and can be None.

        Args:
            datasets (List[str]): [description]

        Raises:
            ValueError: [description]

        Returns:
            Iterable[Tuple[str, str]]: [description]

        Yields:
            Iterator[Iterable[Tuple[str, str]]]: [description]
        """
        for dataset in datasets:
            if dataset.count(":") > 1:
                raise ValueError(f"Invalid dataset '{dataset}'")
            colon_at = dataset.find(":")
            if colon_at != -1:
                yield dataset[:colon_at], dataset[colon_at + 1 :]
            else:
                yield dataset, None

    @property
    def local(self) -> List[Tuple[str, str]]:
        """[summary]

        Returns:
            List[Tuple[str, str]]: [description]
        """
        all_ = self._parse_datasets(self.model.datasets)
        local = self._parse_datasets(self.model.datasets_local)
        return list(all_) + list(local)

    @property
    def cloud(self) -> List[Tuple[str, str]]:
        """[summary]

        Returns:
            List[Tuple[str, str]]: [description]
        """
        all_ = self._parse_datasets(self.model.datasets)
        cloud = self._parse_datasets(self.model.datasets_cloud)
        return list(all_) + list(cloud)


@dataclass
class TrainedModel:
    """TrainedModel Class

    Attributes:
        model: Your model.
        models: A dictionary of models.
        features: A list of the model features.
        params: The parameters of your model.
    """

    # TODO: Simon will look into the field typing
    model: Any = field(default=None)
    models: Dict[str, Any] = field(default_factory=dict)
    features: List[str] = field(default_factory=list)
    params: Dict[str, Any] = field(default_factory=dict)
    validator: Optional[Validator] = None

    class Invalid(Exception):
        """[summary]"""

        pass

    def __new__(cls, **kwargs):
        """Creates a new TrainedModel"""
        if "model" in kwargs and "models" in kwargs:
            raise ValueError(
                (
                    f"Can not instantiate {cls.__name__} using both 'model' and "
                    "'models' parameters. "
                    "Either provide a default model to the 'model' "
                    "parameter, or provide a series of identifiable models "
                    "to the 'models' parameter."
                )
            )
        return object.__new__(cls)

    @property
    def identifiers(self) -> List[str]:
        """[summary]

        Returns:
            List[str]: [description]
        """
        return list(self.models.keys())

    def has_model(self, identifier) -> bool:
        """Check whether a given identifier maps to a Model

        Args:
            identifier ([type]): [description]

        Returns:
            bool: whether or not the identifier maps to a Model
        """
        return identifier in self.models

    def get_model(self, identifier: str = None) -> "Model":
        """Retrieve the model. If an identifier is passed, it retrieves
        the model for that specific identifier.

        Args:
            identifier (str, optional): A unique identifier for the model.
                Defaults to None.

        Raises:
            ValueError: [description]

        """
        if identifier is None:
            return self.get_default_model()
        elif identifier in self.models:
            return self.models[identifier]
        else:
            raise ValueError(f"No model exists with identifier: {identifier}")

    def has_default_model(self) -> bool:
        """Check if the trained model has a model.

        Returns:
            bool: Whether or not the model attribute is None.
        """
        return self.model is not None

    def get_default_model(self):
        """Return the model assosiated with the trained model,
        if one exists.

        """
        if not self.has_default_model():
            raise ValueError(
                "No default model exists for this model. "
                "Use get_model() instead and provide a model identifier."
            )
        return self.model

    # TODO move to function outside class?
    def verify(self):
        """Verify that the TrainedModel object has the correct form."""

        if not isinstance(self.features, list):
            features_type = str(type(self.features))
            raise self.Invalid(
                "Must provide a list of features. "
                f"You gave me something of type {features_type}"
            )
        if not all(isinstance(s, str) for s in self.features):
            raise self.Invalid("All features must be of type str")
        if not [f.strip() for f in self.features if f.strip()]:
            raise self.Invalid(
                (
                    "No feature names provided. "
                    f"Instantiate {self.__class__.__name__} with a list "
                    "of features using the 'features' parameter."
                )
            )


@dataclass
class ModelArtifact:
    """Contains all data that is neccesary for a model to be instanciated and function.
    The artifact is the output of a completion of a model training.

    It contains a TrainedModel and the user-defined code as files.

    It can be used to re-instancesiate a model from an older run,
    to test or deploy the model.

    """

    path: str
    run_id: str
    experiment_name: str
    project_meta: Dict[str, str]
    _tempdir: tempfile.TemporaryDirectory = None

    @property
    def portal_url(self) -> str:
        """The URL to a training run.

        Returns:
            str: The URL of the training run
        """
        return backend.get_portal_url(
            self.project_meta, self.experiment_name, self.run_id
        )

    @classmethod
    def from_cloud(
        cls, experiment_name: str, run_id: str, path=None, project_meta=None
    ) -> "ModelArtifact":
        """Download model and trained_model files from cloud environment
        to a (temporary) folder.

        Args:
            experiment_name (str): The name of the experiment
            run_id (str): The id of the run
            path (str): The path to which the files are downloaded. If None, the files
            are downloaded to a temporary folder.
            project_meta (dict): A dictionary of project information required
            to obtain a workspace. If None, it looks for a project.json
            file with the required information.

        Returns:
            ModelArtifact: This contains model information which makes it possible
            to initiate a LoadedModel

        Example:
            The method can be used as a ContextManager:

            >>> with ModelArtifact.from_cloud(experiment_name, run_id)
            ... as model_artifact:
            >>>     # The downloaded files are downloaded to a temporary folder
            ...     # and they are accesible from the model_artifact-instance
            >>>     print(model_artifact.portal_url)

        It can also be used as an object, but then you have to clean up in the
        temporary folder afterwards:

            >>> model_artifact = ModelArtifact.from_cloud(experiment_name, run_id)
            >>> model_artifact.clean_up() # Clean up in the temporary directory

        """

        if project_meta is None:
            current_path = os.getcwd()
            project_path = locate_file_upwards(current_path, Project._CONFIG_FILE_NAME)

            if project_path is None:
                raise Project.ConfigNotFound(
                    f"Could not find the file '{Project._CONFIG_FILE_NAME}' in this "
                    f"folder (or any of its parents): '{current_path}'.\n"
                    f"I am looking for a folder which contains a file "
                    f"named '{Project._CONFIG_FILE_NAME}' - either in the "
                    f"folder itself or in one of its parent folders. \n\n"
                    f"You can either use the 'project_meta'-argument and provide a "
                    f"dictionary with workspace information, "
                    f"or create a file called '{Project._CONFIG_FILE_NAME}'."
                )

            with open(project_path) as fp:
                project_meta = json.load(fp)

        workspace = backend.get_workspace(project_meta)

        tempdir = None
        if not path:
            tempdir = tempfile.TemporaryDirectory()
            path = os.path.join(tempdir.name, "artifact")

        snapshot_zip_path = backend.download_model_files(
            workspace, experiment_name, run_id, path
        )
        with zipfile.ZipFile(snapshot_zip_path, "r") as zip_ref:
            zip_ref.extractall(path)

        os.remove(snapshot_zip_path)

        artifact = cls(
            path=path,
            run_id=run_id,
            experiment_name=experiment_name,
            project_meta=project_meta,
        )

        if tempdir:
            artifact._tempdir = tempdir

        return artifact

    def clean_up(self):
        """Clean up in temporary directory when using
        :func:`~energinetml.ModelArtifact.from_cloud`
        as a standard function (ie. not a context manager).

        Example:
            >>> model_artifact = ModelArtifact.from_cloud(experiment_name, run_id)
            >>> model_artifact.clean_up() # Clean up in the temporary directory
        """
        if self._tempdir:
            self._tempdir.cleanup()

    def __enter__(self):
        return self

    def __exit__(self, exc, value, tb):
        self.clean_up()


class LoadedModel:
    """Class for a loaded model ready to use in a piece of code.
    The class can be instantiated from a
    :class:`~energinetml.ModelArtifact`-object.

    Attributes:
        model (Model): The model that was loaded.
        trained_model (TrainedModel): The trained model that was loaded.
        controller (PredictionController): Handles the invokation of the user-specified
            prediction logic
    """

    model: "Model"
    trained_model: "TrainedModel"
    controller: PredictionController

    def __init__(self, model: "Model", trained_model: "TrainedModel"):
        """Initialize a LoadedModel

        Args:
            model (Model): The model which holds a
            ``:func: ~energinetml.model.predict``-function
            trained_model (TrainedModel):
        """
        self.model = model
        self.trained_model = trained_model
        self.controller = PredictionController(model, trained_model)

    @classmethod
    def from_artifact(cls, model_artifact: "ModelArtifact") -> "LoadedModel":
        """Creates a LoadedModel from a :class:`~energinetml.ModelArtifact`
        instance.

        Args:
            model_artifact (ModelArtifact): The model artifact that the loaded model
                is created from.

        Returns:
            LoadedModel: A loaded model ready for prediction
        """
        artifact_path = os.path.join(
            model_artifact.path, DEFAULT_RELATIVE_ARTIFACT_PATH
        )
        meta_file_path = os.path.join(
            artifact_path,
            Model._META_FILE_NAME,
        )

        meta_data = cls.read_meta_data(meta_file_path, model_artifact)
        module_name = meta_data["module_name"]

        module_path = os.path.join(
            model_artifact.path, module_name.replace(".", os.sep)
        )

        parent_levels = Model.from_directory(module_path).parent_levels

        model_class = import_model_class(module_path, parent_levels)
        model = model_class.from_directory(module_path)

        trained_model_path = os.path.join(
            artifact_path, model_class._TRAINED_MODEL_FILE_NAME
        )

        trained_model = model.load(trained_model_path)

        return cls(model, trained_model)

    def predict(self, input_data: Union[pd.DataFrame, str]) -> PredictResponse:
        """Make predictions based on data from a DataFrame or a json string.
        The prediction logic is specified in the user-defined
        :func:`~energinetml.Model.predict`.

        Args:
            input_data (Union[pd.DataFrame, str]): Either a pd.DataFrame
            or a json string

        Returns:
            PredictResponse: A pydantic model that holds predictions and validations
        """

        if isinstance(input_data, pd.DataFrame):
            # Convert pd.DataFrame to json
            json_data = df_to_pydantic_json(input_data)
        else:
            # The data is already in json format
            json_data = input_data

        # Load data model from JSON
        try:
            request = pydantic.parse_obj_as(self.controller.request_model, json_data)
        except pydantic.error_wrappers.ValidationError as e:
            print("Invalid input JSON provided. Error description follows:")
            print(str(e.json(indent=4)))

        # Invoke user-defined prediction logic
        try:
            _, _, predictions, validations = self.controller.invoke_model(request)
        except NotImplementedError:
            print("Prediction script needs an implementation!")
            print(
                "The predict() method of your model raised a NotImplementedError "
                "which indicates that you have not yet implemented it."
            )
            print("Stacktrace follows:")
            print("-" * 79)
            raise

        return self.controller.response_model(
            predictions=predictions, validations=validations
        )

    @staticmethod
    def read_meta_data(meta_file_path: str, model_artifact: "ModelArtifact") -> Dict:
        """Read meta data from a filepath to a dictionary.

        Args:
            meta_file_path (str): The filepath to the metadata file
            model_artifact (ModelArtifact): Used for error message

        Raises:
            MetaFileMissing: Thrown if the metafile doesn't exit at the filepath

        Returns:
            Dict: Dictionary with the metadata
        """
        try:
            with open(meta_file_path) as meta_data_file:
                meta_data = json.load(meta_data_file)
        except FileNotFoundError:
            raise MetaFileMissing(model_artifact)

        return meta_data


# -- Model importing ---------------------------------------------------------


class MetaFileMissing(Exception):
    def __init__(self, model_artifact):
        msg = (
            f"Could not find a meta data file "
            f"'{Model._META_FILE_NAME}' in the experiment folder "
            f"'{DEFAULT_RELATIVE_ARTIFACT_PATH}' from the run\n"
            f"Portal link: {model_artifact.portal_url}\n\n"
            f"Check that the file '{Model._META_FILE_NAME}' exists. If not, "
            "the experiment may have been run with an older version of the SDK."
        )

        super().__init__(msg)


class ModelError(Exception):
    """[summary]"""

    pass


class ModelModuleError(ModelError):
    """
    Raised if there is a problem importing the usermodel as a module.
    """

    def __init__(self, msg):
        user_msg = """\n
            There was a problem importing modules in your 'model.py'.\n
            If you import a python module from a parent folder, remember to add
            the relative filepath to 'files_include' in 'model.json."""
        output_msg = msg + user_msg

        super().__init__(output_msg)


class ModelImportError(ModelError):
    """
    Raised if script does not contain a 'model' object
    in the global scope.
    """

    pass


class ModelNotClassError(ModelError):
    """
    Raised if imported 'model' object is not a class type.
    """

    pass


class ModelNotInheritModel(ModelError):
    """
    Raised if imported 'model' does not inherit from Model.
    """

    pass


def import_model_class(path, parent_levels):
    """
    Imports user defined 'model' object from python-module at 'path'.
    Validates that its a class, and that it inherits from Model.

    Args:
        path (str): path to the Model class that should be imported.
        parent_levels (int):
            How many parent levels to go up in the path to support
            relative imports.

    Raises:
        ModelImportError: [description]
        ModelNotClassError: [description]
        ModelNotInheritModel: [description]

    Returns:
        Model: A User defined Model class
    """
    modules = []
    folder = path

    # Go up in the path to support relative imports
    for _ in range(parent_levels):
        folder, name = os.path.split(os.path.normpath(folder))
        modules.append(name)

    if folder not in sys.path:
        sys.path.append(folder)

    # Revert list of module names to get python module path
    # This is to support relative imports in model.py
    module_to_import = ".".join(modules[::-1])

    try:
        module = importlib.import_module(module_to_import)
    except ImportError as e:
        raise ModelModuleError(e.msg)

    if not hasattr(module, "model"):
        raise ModelImportError(module)

    sys.modules["model"] = module

    model_class = getattr(module, "model")

    if not inspect.isclass(model_class):
        raise ModelNotClassError()
    if not issubclass(model_class, Model):
        raise ModelNotInheritModel()

    return model_class


def df_to_pydantic_json(df: pd.DataFrame) -> Dict[str, List[Any]]:
    """Convert a pd.DataFrame object to a dictionary in expected format.

    Args:
        df (pd.DataFrame): _description_

    Returns:
        Dict[str, List[Any]]: _description_
    """
    json_dict = df.to_dict(orient="records")
    features_list = []
    for record in json_dict:
        features_list.append({"features": record})

    json_data = {"inputs": features_list}
    return json_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Manage machine learning models.
"""
import click

from .build import build as build_model
from .files import files as model_files
from .init import init_model
from .release import release as release_model
from .serve import serve as serve_model
from .submit import submit as submit_model
from .train import train as train_model


@click.group()
def model_group():
    """Manage machine learning models."""
    pass


model_group.add_command(init_model, "init")
model_group.add_command(train_model)
model_group.add_command(build_model)
model_group.add_command(serve_model)
model_group.add_command(submit_model)
model_group.add_command(release_model)
model_group.add_command(model_files)



import json
import os
import tempfile

import pytest

from energinetml.core.project import MachineLearningProject
from energinetml.settings import PACKAGE_NAME, PACKAGE_VERSION

NAME = "NAME"
SUBSCRIPTION_ID = "SUBSCRIPTION-ID"
RESOURCE_GROUP = "RESOURCE-GROUP"
WORKSPACE_NAME = "WORKSPACE-NAME"
VNET = "VNET"
SUBNET = "SUBNET"


@pytest.fixture
def project():
    with tempfile.TemporaryDirectory() as path:
        yield MachineLearningProject(
            path=path,
            name=NAME,
            subscription_id=SUBSCRIPTION_ID,
            resource_group=RESOURCE_GROUP,
            workspace_name=WORKSPACE_NAME,
            vnet_name=VNET,
            subnet_name=SUBNET,
        )


class TestProject:
    def test__vnet_resourcegroup_name(self, project):
        """
        :param Project project:
        """
        assert project.vnet_resourcegroup_name == RESOURCE_GROUP

    def test__vnet_name(self, project):
        """
        :param Project project:
        """
        assert project.vnet_name == VNET

    def test__subnet_name(self, project):
        """
        :param Project project:
        """
        assert project.subnet_name == SUBNET

    def test__requirements_file_path(self, project):
        """
        :param Project project:
        """
        assert project.requirements_file_path == os.path.join(
            project.path, "requirements.txt"
        )

    def test__default_model_path(self, project):
        """
        :param Project project:
        """
        model_name = "my-model"

        assert project.default_model_path(model_name) == os.path.join(
            project.path, model_name
        )

    def test__create__should_create_project_files(self):
        with tempfile.TemporaryDirectory() as path:
            project = MachineLearningProject.create(
                path=path,
                name=NAME,
                subscription_id=SUBSCRIPTION_ID,
                resource_group=RESOURCE_GROUP,
                workspace_name=WORKSPACE_NAME,
                vnet_name=VNET,
                subnet_name=SUBNET,
            )

            assert os.path.isfile(os.path.join(project.path, "project.json"))
            assert os.path.isfile(os.path.join(project.path, "requirements.txt"))

            # requirements.txt
            assert PACKAGE_NAME in project.requirements
            assert project.requirements.get(PACKAGE_NAME).specs == [
                ("==", str(PACKAGE_VERSION))
            ]  # noqa: E501

            # project.json
            with open(os.path.join(project.path, "project.json")) as f:
                config = json.load(f)
                assert config["name"] == NAME
                assert config["subscription_id"] == SUBSCRIPTION_ID
                assert config["resource_group"] == RESOURCE_GROUP
                assert config["workspace_name"] == WORKSPACE_NAME



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Energinet ML python module
"""


import logging

logging.getLogger("opentelemetry.trace.status").setLevel(logging.ERROR)


from .settings import (  # noqa: E402
    APPINSIGHTS_INSTRUMENTATIONKEY,
    PACKAGE_NAME,
    PACKAGE_VERSION,
)

__name__ = PACKAGE_NAME
__version__ = PACKAGE_VERSION


# OpenTelemetry must be configured before importing any packages using it
if APPINSIGHTS_INSTRUMENTATIONKEY:

    from azure.monitor.opentelemetry.exporter import AzureMonitorTraceExporter
    from opentelemetry import trace
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import BatchSpanProcessor
    from opentelemetry.sdk.trace.sampling import ALWAYS_ON  # noqa: E402

    conn = f"InstrumentationKey={APPINSIGHTS_INSTRUMENTATIONKEY}"
    exporter = AzureMonitorTraceExporter(connection_string=conn)
    span_processor = BatchSpanProcessor(exporter)

    trace.set_tracer_provider(TracerProvider(sampler=ALWAYS_ON))
    provider = trace.get_tracer_provider()
    provider.add_span_processor(span_processor)


# Importing pandas must be done before importing modules using OpenTelemetry,
# otherwise an (apparent) bug in Pandas causes an exception in a second thread
import pandas  # noqa: E402

from .cli import main  # noqa: E402
from .core.insight import (  # noqa: E402
    query_predictions,
    query_predictions_as_dataframe,
)
from .core.logger import MetricsLogger  # noqa: E402
from .core.model import LoadedModel, Model, ModelArtifact, TrainedModel  # noqa: E402
from .core.predicting import PredictionInput  # noqa: E402
from .core.project import Project  # noqa: E402
from .core.training import requires_parameter  # noqa: E402



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import click

from energinetml.cli.utils import discover_project
from energinetml.core.docker import build_webapp_docker_image
from energinetml.core.project import WebAppProject


@click.command()
@click.option(
    "--tag",
    "-t",
    required=True,
    help="Name and optionally a tag in the `name:tag` format",
)
@discover_project(WebAppProject)
def build(tag: str, project: WebAppProject) -> None:
    """Build web app Docker image."""
    build_webapp_docker_image(project=project, tag=tag)



import typing
from enum import Enum
from unittest.mock import ANY, Mock, PropertyMock, patch

import pydantic
import pytest
from fastapi.testclient import TestClient

from energinetml.core.http import ModelMetaData, create_app, run_predict_api
from energinetml.settings import PACKAGE_REQUIREMENT

MODEL_NAME = "model-name"
MODEL_VERSION = "123"
MODEL_EXPERIMENT = "model-experiment"
MODEL_VALIDATOR = True
TEST_SERVER_BASE_URL = "https://testserver"
MODEL_DATASETS = ["dataset1", "dataset2"]
SUBSCRIPTION_ID = "subscription-id"
RESOURCE_GROUP = "resource-group"
WORKSPACE_NAME = "workspace-name"
RUN_ID = "run-id"
PORTAL_URL = "portal-url"
PORT = 5678
HOST = "host"

PREDICT_ENDPOINT = "/predict"
HEALTH_ENDPOINT = "/health"
METADATA_ENDPOINT = "/metadata"


# -- Request & Response models -----------------------------------------------


class IdentifierEnum(Enum):
    identifier1 = "identifier1"
    identifier2 = "identifier2"


class PredictFeatures(pydantic.BaseModel):
    feature1: int
    feature2: int


class PredictInput(pydantic.BaseModel):
    identifier: IdentifierEnum
    features: PredictFeatures


class PredictRequest(pydantic.BaseModel):
    inputs: typing.List[PredictInput]


class PredictResponse(pydantic.BaseModel):
    predictions: typing.List[typing.Any]


# -- Fixtures ----------------------------------------------------------------


def create_test_models():
    model = Mock()
    type(model).name = PropertyMock(return_value=MODEL_NAME)
    type(model).experiment = PropertyMock(return_value=MODEL_EXPERIMENT)
    type(model).datasets = PropertyMock(return_value=MODEL_DATASETS)

    trained_model = Mock()
    type(trained_model).validator = PropertyMock(return_value=MODEL_VALIDATOR)

    model_meta_data = ModelMetaData(
        model_version=MODEL_VERSION,
        model_name=model.name,
        model_experiment_id=model.experiment,
        has_validator=True if trained_model.validator else False,
        datasets=MODEL_DATASETS,
        subscription_id=SUBSCRIPTION_ID,
        resource_group=RESOURCE_GROUP,
        workspace_name=WORKSPACE_NAME,
        run_id=RUN_ID,
        portal_url=PORTAL_URL,
    )

    meta_data = {
        "run_id": RUN_ID,
        "subscription_id": SUBSCRIPTION_ID,
        "resource_group": RESOURCE_GROUP,
        "workspace_name": WORKSPACE_NAME,
        "portal_url": PORTAL_URL,
    }

    return model, trained_model, model_meta_data, meta_data


@pytest.fixture
def controller():
    yield Mock(request_model=PredictRequest, response_model=PredictResponse)


@pytest.fixture
def client(controller):
    with patch(
        "energinetml.core.http.PredictionController"
    ) as PredictionController_mock:  # noqa: E501

        PredictionController_mock.return_value = controller

        model, trained_model, model_meta_data, _ = create_test_models()

        app = create_app(
            model=model,
            trained_model=trained_model,
            model_meta_data=model_meta_data,
            model_version=MODEL_VERSION,
        )

        yield TestClient(app)


@pytest.fixture
def client_with_opentelemetry(controller):
    with patch(
        "energinetml.core.http.PredictionController"
    ) as PredictionController_mock, patch(
        "energinetml.core.http.APPINSIGHTS_INSTRUMENTATIONKEY"
    ):

        PredictionController_mock.return_value = controller

        model, trained_model, model_meta_data, _ = create_test_models()

        app = create_app(
            model=model,
            trained_model=trained_model,
            model_version=MODEL_VERSION,
            model_meta_data=model_meta_data,
        )

        yield TestClient(app=app, base_url=TEST_SERVER_BASE_URL)


# -- health_http_endpoint() Tests --------------------------------------------


def test__health_http_endpoint__should_return_status_200(client):
    """
    :param TestClient client:
    """
    # Act
    response = client.get(HEALTH_ENDPOINT)

    # Assert
    assert response.status_code == 200


# -- metadata_http_endpoint() Tests --------------------------------------------


def test__metadata_http_endpoint__should_return_status_200_and_body(client):
    """
    :param TestClient client:
    """
    # Act
    response = client.get(METADATA_ENDPOINT)
    m = pydantic.parse_obj_as(ModelMetaData, response.json())

    # Assert
    assert response.status_code == 200

    assert m.model_version == MODEL_VERSION
    assert m.model_name == MODEL_NAME
    assert m.model_experiment_id == MODEL_EXPERIMENT
    assert m.datasets == MODEL_DATASETS
    assert m.has_validator == MODEL_VALIDATOR
    assert m.subscription_id == SUBSCRIPTION_ID
    assert m.resource_group == RESOURCE_GROUP
    assert m.workspace_name == WORKSPACE_NAME
    assert m.run_id == RUN_ID
    assert m.portal_url == PORTAL_URL


def test__metadata_http_endpoint__should_return_status_200(client):
    """
    :param TestClient client:
    """
    # Act
    response = client.get(METADATA_ENDPOINT)

    # Assert
    assert response.status_code == 200


# -- predict_http_endpoint() Tests -------------------------------------------


def test__predict_http_endpoint__omit_inputs__should_return_status_422(
    client, controller
):
    """
    :param TestClient client:
    :param Mock controller:
    """
    # Act
    response = client.post(PREDICT_ENDPOINT)

    # Assert
    assert response.status_code == 422
    controller.predict.assert_not_called()


def test__predict_http_endpoint__omit_identifier__should_return_status_422(
    client, controller
):
    """
    :param TestClient client:
    :param Mock controller:
    """
    # Act
    response = client.post(
        PREDICT_ENDPOINT,
        json={"inputs": [{"features": {"feature1": 1, "feature2": 2}}]},
    )

    # Assert
    assert response.status_code == 422
    controller.predict.assert_not_called()


def test__predict_http_endpoint__omit_feature__should_return_status_422(
    client, controller
):
    """
    :param TestClient client:
    :param Mock controller:
    """
    # Act
    response = client.post(
        PREDICT_ENDPOINT,
        json={
            "inputs": [
                {
                    "identifier": "identifier1",
                    "features": {
                        "feature1": 1,
                        # feature2 missing
                    },
                }
            ]
        },
    )

    # Assert
    assert response.status_code == 422
    controller.predict.assert_not_called()


def test__predict_http_endpoint__should_return_status_200(client, controller):
    """
    :param TestClient client:
    :param Mock controller:
    """
    controller.predict.return_value = PredictResponse(predictions=[1, 2])

    # Act
    response = client.post(
        PREDICT_ENDPOINT,
        json={
            "inputs": [
                {
                    "identifier": "identifier1",
                    "features": {"feature1": 1, "feature2": 2},
                }
            ]
        },
    )

    # Assert
    assert response.status_code == 200
    assert response.headers["X-sdk-version"] == str(PACKAGE_REQUIREMENT)
    assert response.json() == {"predictions": [1, 2]}

    controller.predict.assert_called_once()


# -- OpenTelemetry middleware Tests ------------------------------------------


@patch("energinetml.core.http.tracer")
def test__opentelemetry_middleware__predict_raised_exception__should_log_exception(
    tracer_mock, client_with_opentelemetry, controller
):
    """
    :param Mock tracer_mock:
    :param TestClient client_with_opentelemetry:
    :param Mock controller:
    """
    controller.predict.side_effect = RuntimeError

    span = Mock()
    tracer_mock.start_span.return_value.__enter__.return_value = span

    # Act
    response = client_with_opentelemetry.post(
        PREDICT_ENDPOINT,
        json={
            "inputs": [
                {
                    "identifier": "identifier1",
                    "features": {"feature1": 1, "feature2": 2},
                }
            ]
        },
    )

    # Assert
    assert response.status_code == 500
    controller.predict.assert_called_once()

    # Common attributes
    span.set_attribute.assert_any_call("http.url", f"{TEST_SERVER_BASE_URL}/predict")
    span.set_attribute.assert_any_call("http_url", f"{TEST_SERVER_BASE_URL}/predict")
    span.set_attribute.assert_any_call("model_name", MODEL_NAME)
    span.set_attribute.assert_any_call("model_version", MODEL_VERSION)

    # Error attributes
    span.record_exception.assert_called_once()
    span.set_status.assert_called_once()
    span.set_attribute.assert_any_call("http.status_code", 500)
    span.set_attribute.assert_any_call("http_status_code", 500)
    span.set_attribute.assert_any_call("error.name", ANY)
    span.set_attribute.assert_any_call("error.message", ANY)
    span.set_attribute.assert_any_call("error.stacktrace", ANY)


@patch("energinetml.core.http.tracer")
def test__opentelemetry_middleware__predict_ok__should_log_status_code(
    tracer_mock, client_with_opentelemetry, controller
):
    """
    :param Mock tracer_mock:
    :param TestClient client_with_opentelemetry:
    :param Mock controller:
    """
    controller.predict.return_value = PredictResponse(predictions=[1, 2])

    span = Mock()
    tracer_mock.start_span.return_value.__enter__.return_value = span

    # Act
    response = client_with_opentelemetry.post(
        PREDICT_ENDPOINT,
        json={
            "inputs": [
                {
                    "identifier": "identifier1",
                    "features": {"feature1": 1, "feature2": 2},
                }
            ]
        },
    )

    # Assert
    assert response.status_code == 200
    controller.predict.assert_called_once()

    # Common attributes
    span.set_attribute.assert_any_call("http.url", f"{TEST_SERVER_BASE_URL}/predict")
    span.set_attribute.assert_any_call("http_url", f"{TEST_SERVER_BASE_URL}/predict")
    span.set_attribute.assert_any_call("model_name", MODEL_NAME)
    span.set_attribute.assert_any_call("model_version", MODEL_VERSION)

    # Success attributes
    span.set_status.assert_called_once()
    span.set_attribute.assert_any_call("http.status_code", 200)
    span.set_attribute.assert_any_call("http_status_code", 200)


# -- run_predict_api() Tests -------------------------------------------------


@patch("energinetml.core.http.create_app")
@patch("energinetml.core.http.uvicorn.run")
def test__run_predict_api__should_create_and_run_app(run_mock, create_app_mock):
    """
    :param Mock run_mock:
    :param Mock create_app_mock:
    """

    model, trained_model, model_meta_data, meta_data = create_test_models()

    app = Mock()

    create_app_mock.return_value = app

    # Act
    run_predict_api(
        model=model,
        trained_model=trained_model,
        meta_data=meta_data,
        model_version=MODEL_VERSION,
        host=HOST,
        port=PORT,
    )

    # Assert
    create_app_mock.assert_called_once_with(
        model=model,
        trained_model=trained_model,
        model_version=MODEL_VERSION,
        model_meta_data=model_meta_data,
    )

    run_mock.assert_called_once_with(app=app, host=HOST, port=PORT)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import contextlib
import glob
import os
import shutil
import tempfile
from typing import ContextManager, Iterable, List, Set, Tuple


class FileMatcher:
    """[summary]"""

    def __init__(
        self,
        root_path: str,
        include: List[str],
        exclude: List[str] = None,
        recursive: bool = True,
    ):
        """[summary]

        Args:
            root_path (str): [description]
            include (List[str]): [description]
            exclude (List[str], optional): [description]. Defaults to None.
            recursive (bool, optional): [description]. Defaults to True.
        """
        self.root_path = root_path
        self.include = include
        self.exclude = exclude
        self.recursive = recursive

    def __iter__(self) -> Iterable[str]:
        """[summary]

        Returns:
            Iterable[str]: [description]
        """
        include_files = self.match_files(self.include)
        exclude_files = (
            self.match_files(self.exclude) if self.exclude is not None else set()
        )

        return iter(include_files - exclude_files)

    def match_files(self, patterns: List[str]) -> Set[str]:
        """[summary]

        Args:
            patterns (List[str]): [description]

        Returns:
            Set[str]: [description]
        """
        all_matches = set()

        for pattern in patterns:
            matches = glob.iglob(
                pathname=f"{self.root_path}/{pattern}", recursive=self.recursive
            )

            all_matches.update(
                os.path.relpath(path, self.root_path)
                for path in matches
                if os.path.isfile(path)
            )

        return all_matches


@contextlib.contextmanager
def temporary_folder(files: Iterable[Tuple[str, str]]) -> ContextManager[None]:
    """Returns a context manager which creates a temporary folder on the
    filesystem and copies files into the folder.

    The `files` parameter is an iterable of tuples of (src, dst), where
    src is absolute path on the filesystem (a file to copy), and dst
    is the destination in the temporary folder relative to its root.

    Usage example:

        files_to_copy = [
            ('/home/folder/some-file.txt', 'some-file.txt'),
            ('/home/folder/another-file.txt', 'folder/some-file.txt'),
        ]

        with temporary_folder() as temp_path:
            # files are available in temp_path

    Args:
        files (Iterable[Tuple[str, str]]): [description]

    Returns:
        Generator[str]: [description]
    """

    def __mkdir_and_copyfile(src, dst):
        if not os.path.isdir(os.path.split(dst)[0]):
            os.makedirs(os.path.split(dst)[0])
        shutil.copyfile(src, dst)

    with tempfile.TemporaryDirectory() as temp_folder_path:
        for src, dst in files:
            __mkdir_and_copyfile(src, os.path.join(temp_folder_path, dst))
        yield temp_folder_path



import requests


def test__deployment__health(deployment_base_url):
    """
    Invokes model health endpoint.

    :param str deployment_base_url:
    """

    # -- Act -----------------------------------------------------------------

    response = requests.get(url=f"{deployment_base_url}/health")

    # -- Assert --------------------------------------------------------------

    assert response.status_code == 200


def test__deployment__model_predict(
    deployment_base_url, prediction_input, prediction_output
):
    """
    Invokes prediction on a model using the provided base URL.

    :param str deployment_base_url:
    :param typing.Any prediction_input:
    :param typing.Any prediction_output:
    """

    # -- Act -----------------------------------------------------------------

    response = requests.post(
        url=f"{deployment_base_url}/predict", json=prediction_input
    )

    # -- Assert --------------------------------------------------------------

    assert response.status_code == 200
    assert response.json()["predictions"] == prediction_output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]"""
from __future__ import annotations

import os
import subprocess
import sys
from typing import TYPE_CHECKING  # noqa TYP001

if TYPE_CHECKING:

    from energinetml.core.project import Project


class WebAppEngine:
    """[summary]"""

    command = None

    def serve(self, project: Project) -> None:
        """[summary]

        Args:
            project (Project): [description]

        Raises:
            NotImplementedError: [description]
        """
        os.chdir(os.path.join(project.path, "src"))

        if self.command is None:
            raise NotImplementedError

        subprocess.check_call(
            self.command, stdout=sys.stdout, stderr=subprocess.STDOUT, shell=False
        )


class WsgiWebApp(WebAppEngine):
    """
    WSGI
    """

    command = ("waitress-serve", "--listen=*:8000", "app:app")


class AsgiWebApp(WebAppEngine):
    """
    ASGI
    """

    command = ("uvicorn", "app:app")


class OpyratorWebApp(WebAppEngine):
    """[summary]"""

    def serve(self, path: str):
        """[summary]

        Args:
            path (str): [description]

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
from __future__ import annotations

import os
from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from energinetml import Project

from energinetml.core.model import (
    Model,
    ModelImportError,
    ModelNotClassError,
    ModelNotInheritModel,
    import_model_class,
)


def discover_project(project_cls: Project, required: bool = True) -> str:
    """[summary]

    Args:
        project_cls (Project): [description]
        required (bool, optional): [description]. Defaults to True.

    Returns:
        str: [description]
    """

    def _project_from_path_callback(
        ctx: click.Context, param: click.Parameter, value: str
    ) -> str:
        """[summary]

        Args:
            ctx (click.Context): [description]
            param (click.Parameter): [description]
            value (str): [description]

        Raises:
            click.Abort: [description]

        Returns:
            str: [description]
        """
        try:
            return project_cls.from_directory(value)
        except project_cls.ConfigNotFound:
            if required:
                click.echo(
                    (
                        "Could not find a project in this folder "
                        f"(or any of its parents): {value}"
                    )
                )

                click.echo(
                    (
                        "I am looking for a folder which contains a file "
                        f"named '{project_cls._CONFIG_FILE_NAME}' - either in the "
                        "folder itself or in one of its parent folders."
                    )
                )

                click.echo(
                    "Specify which project to use by providing the "
                    "-p/--path parameter."
                )

                raise click.Abort()

    return click.option(
        "--path",
        "-p",
        "project",
        default=".",
        type=click.Path(dir_okay=True, resolve_path=True),
        help="Project directory path (default: current directory)",
        callback=_project_from_path_callback,
    )


def discover_model(
    required: bool = True, load_model: bool = True, param_name: str = "model"
) -> None:
    """[summary]

    Args:
        required (bool, optional): [description]. Defaults to True.
        load_model (bool, optional): [description]. Defaults to True.
        param_name (str, optional): [description]. Defaults to "model".
    """

    def _model_from_path_callback(
        ctx: click.Context, param: click.Parameter, model_file_path: str
    ) -> Model:
        """[summary]

        Args:
            ctx (click.Context): [description]
            param (click.Parameter): [description]
            value (str): [description]

        Raises:
            click.Abort: [description]

        Returns:
            Model: [description]
        """
        # Does model module exist?
        if not os.path.isdir(model_file_path):
            click.echo(("Could not find this folder: %s") % model_file_path)

            click.echo(("I am looking for a folder which contains a python module."))

            click.echo(
                ("Specify which model to use by providing the " "-p/--path parameter.")
            )

            raise click.Abort()

        parent_levels = Model.from_directory(model_file_path).parent_levels

        try:
            model_class = import_model_class(model_file_path, parent_levels)
        except ModelImportError:
            # Imported script does not have a "model" attribute
            click.echo(
                f"Failed to import your model class from module: {model_file_path}"
            )

            click.echo(
                'Make sure you refer your model class and name it "model". '
                'Do this by defining a global variable named "model" in your '
                "module, and point it to your model class."
            )

            click.echo("")
            click.echo("Example:")
            click.echo("")
            click.echo("    class MyModel(Model):")
            click.echo("        ...")
            click.echo("")
            click.echo("    model = MyModel")
            click.echo("")

            raise click.Abort()
        except ModelNotClassError:
            # Imported "model" attribute is not a Class type
            click.echo(
                f"Failed to import your model class from module: {model_file_path}"
            )

            click.echo(
                'When you define the "model" object in your model script, '
                "make sure not to instantiate it."
            )

            click.echo("")
            click.echo("Example of doing it correct:")
            click.echo("")
            click.echo("    class MyModel(Model):")
            click.echo("        ...")
            click.echo("")
            click.echo("    model = MyModel")

            click.echo("")
            click.echo("Example of doing it WRONG:")
            click.echo("")
            click.echo("    class MyModel(Model):")
            click.echo("        ...")
            click.echo("")
            click.echo("    model = MyModel()  # Notice the instantiation")
            click.echo("")

            raise click.Abort()
        except ModelNotInheritModel:
            # Imported "model" attribute does not inherit from Model
            click.echo(
                f"Failed to import your model class from module: {model_file_path}"
            )

            click.echo("The model you are referring to does not inherit from Model.")

            click.echo("")
            click.echo("Example of doing it correct:")
            click.echo("")
            click.echo("    from energinetml import Model")
            click.echo("")
            click.echo("    class MyModel(Model):  # Notice the inheritance")
            click.echo("        ...")
            click.echo("")
            click.echo("    model = MyModel")
            click.echo("")

            raise click.Abort()

        try:
            return model_class.from_directory(model_file_path)
        except Model.ConfigNotFound:
            if required:
                click.echo(
                    (
                        "Could not find a configuration in this folder "
                        f"(or any of its parents): {model_file_path}"
                    )
                )

                click.echo(
                    (
                        "I am looking for a folder which contains a file "
                        f"named {Model._CONFIG_FILE_NAME} - either in the "
                        "folder itself or in one of its parent folders."
                    )
                )

                click.echo(
                    (
                        "Specify which model to use by providing the "
                        "-p/--path parameter."
                    )
                )

                raise click.Abort()

    return click.option(
        "--path",
        "-p",
        "model",
        default=".",
        type=click.Path(dir_okay=True, resolve_path=True),
        help="Model directory path (default: current directory)",
        callback=_model_from_path_callback,
    )



from unittest.mock import Mock

from energinetml.core.model import TrainedModel
from energinetml.core.predicting import PredictionController


class TestPredictionController:
    def test__predictions_returned_in_correct_order(self):
        def __model_predict(trained_model, identifier, input_data):
            """
            :param TrainedModel trained_model:
            :param str identifier:
            :param typing.List[typing.Dict[str, typing.Any]] input_data:
            """
            return [identifier for _ in range(len(input_data))]

        trained_model = TrainedModel(
            features=["feature1", "feature2"],
            models={"identifier1": Mock(), "identifier2": Mock()},
        )

        uut = PredictionController(
            model=Mock(predict=__model_predict),
            trained_model=trained_model,
            model_version="123",
        )

        request = uut.request_model(
            inputs=[
                {
                    "identifier": "identifier1",
                    "features": {"feature1": 1, "feature2": 1},
                },
                {
                    "identifier": "identifier2",
                    "features": {"feature1": 2, "feature2": 2},
                },
                {
                    "identifier": "identifier1",
                    "features": {"feature1": 3, "feature2": 3},
                },
            ]
        )

        # Act
        result = uut.predict(request, correlation_id="uuid")

        # Assert
        assert result.predictions == ["identifier1", "identifier2", "identifier1"]
        assert result.validations is None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

import json
import os
from dataclasses import asdict, dataclass
from typing import Dict, Union

from energinetml.settings import DEFAULT_ENCODING


@dataclass
class Configurable:
    """[summary]"""

    class ConfigNotFound(Exception):
        """[summary]"""

        pass

    class ConfigError(Exception):
        """_summary_

        Args:
            Exception (_type_): _description_

        Raises:
            RuntimeError: _description_
            cls.NotFound: _description_

        Returns:
            _type_: _description_
        """

        pass

    # Constants
    _CONFIG_FILE_NAME = None

    # Members
    path: str

    @classmethod
    def create(cls, path: str, **kwargs) -> "Configurable":
        """[summary]

        Args:
            path (str): [description]

        Returns:
            Configurable: [description]
        """
        obj = cls(path=path, **kwargs)
        obj.save()
        return obj

    @classmethod
    def from_config_file(cls, file_path: str) -> "Configurable":
        """[summary]

        Args:
            file_path (str): [description]

        Returns:
            Configurable: [description]
        """
        with open(file_path, encoding=DEFAULT_ENCODING) as f:
            return cls(path=os.path.split(file_path)[0], **json.load(f))

    @classmethod
    def from_directory(cls, path: str) -> "Configurable":
        """[summary]

        Args:
            path (str): [description]

        Raises:
            RuntimeError: [description]
            cls.NotFound: [description]

        Returns:
            Configurable: [description]
        """
        if cls._CONFIG_FILE_NAME is None:
            raise RuntimeError("Attribute _CONFIG_FILE_NAME is None")

        file_pointer = locate_file_upwards(path, cls._CONFIG_FILE_NAME)

        if file_pointer is not None:
            return cls.from_config_file(file_pointer)
        else:
            raise cls.ConfigNotFound()

    def get_file_path(self, *relative_path: str) -> str:
        """Returns absolute path to a file at relative_path,
        where relative_path is relative to config file.

        Args:
            *relative_path (str): [description]

        Returns:
            str: [description]
        """

        return os.path.abspath(os.path.join(self.path, *relative_path))

    def get_relative_file_path(self, absolute_path: str) -> str:
        """Provided an absolute file path, returns the path relative
        to config file.

        Args:
            absolute_path (str): [description]

        Returns:
            str: [description]
        """
        return os.path.relpath(absolute_path, self.path)

    def save(self) -> None:
        """Saved config as JSON to filesystem."""
        if not os.path.isdir(self.path):
            os.makedirs(self.path)

        config_file_path = self.get_file_path(self._CONFIG_FILE_NAME)
        with open(config_file_path, "w", encoding=DEFAULT_ENCODING) as f:
            d = asdict(self)
            d.pop("path")
            json.dump(d, f, indent=4, sort_keys=True)

    def as_dict(self) -> Dict[str, str]:
        """Get the object variables as a dictionary.
        Helpful in creation of a meta_data dictionary

        returns:
            Dict[str, str]:
        """
        return self.__dict__


def locate_file_upwards(path: str, filename: str) -> Union[str, None]:
    """locate a file and return its path. starts from the path-parameter and searches
    upwards to the root.

    args:
        path (str): the path where the search will start
        filename (str): the filename of the file to search for

    returns:
        str: the path of the file if found
    """

    def __is_root(_path: str) -> str:
        """[summary]

        args:
            _path (str): [description]

        returns:
            str: [description]
        """
        # you have yourself root.
        # works on windows and *nix paths.
        # does not work on windows shares (\\server\share)
        return os.path.dirname(_path) == _path

    while 1:
        file_pointer = os.path.join(path, filename)
        if os.path.isfile(file_pointer):
            return file_pointer
        elif __is_root(path):
            return None
        else:
            path = os.path.abspath(os.path.join(path, ".."))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import io
import os
from typing import TYPE_CHECKING, Dict, Iterable, List, Tuple, Union

from azureml import exceptions
from azureml.core import Dataset, Workspace
from azureml.data import FileDataset, TabularDataset
from pandas import DataFrame

from energinetml.settings import DEFAULT_ENCODING

if TYPE_CHECKING:
    from energinetml.core.model import Model


class MLDataSet:
    """[summary]"""

    def __init__(
        self, name: str, mount_path: str = None, dataframe: DataFrame = None
    ) -> None:
        """[summary]

        Args:
            name (str): [description]
            mount_path (str): [description]
            dataframe (DataFrame, optional): Dataframe of the dataset.
            This is set if the dataset is tabular. Defaults to None.
        """
        self.name = name
        self.mount_path = mount_path
        self.dataframe = dataframe

    def __str__(self) -> str:
        """[summary]

        Returns:
            str: [description]
        """
        return f"{self.__class__.__name__}<{self.name}>"

    def path(self, *relative_path: Union[List[str], str]) -> str:
        """[summary]

        Args:
            *relative_path (Union[List[str], str]): [description]

        Returns:
            str: [description]
        """
        return os.path.join(self.mount_path, *relative_path)

    def open(self, relative_path: List[str], *args, **kwargs) -> io.TextIOWrapper:
        """
        :param list[str] relative_path:
        :param args: *args for open()
        :param kwargs: **kwargs for open()
        :rtype: typing.IO
        """
        return open(
            self.path(*relative_path), *args, encoding=DEFAULT_ENCODING, **kwargs
        )

    def contains(self, *relative_path: Union[List[str], str]) -> bool:
        """[summary]

        Args:
            *relative_path (Union[List[str], str]): [description]

        Returns:
            bool: [description]
        """
        return os.path.exists(self.path(*relative_path))


class MLDataStore(Dict[str, "MLDataSet"]):
    """A class that works as a dictionary of registered datasets in
    :func:`~energinetml.Model.train`.

    Example:
        When in the :func:`~energinetml.Model.train`-function,
        you can acces a dataset in this way:
        .. code-block:: python

            # We access the "iris" dataset using its name as key:
            iris_dataset = datasets['iris']

            # To access files within the iris dataset, we can use the path() method,
            # which returns an absolute path to a file in the dataset.
            # In this case, the dataset has a "iris.csv" file:
            iris_file_path = iris_dataset.path('iris.csv')

            # Be aware that accessing files within sub-folders should be done
            # using the same approach as when using os.path.join() to allow
            # compatibility across different operating systems.
            # The following returns the path to file "some-folder/iris2.csv":
            iris_file_path2 = iris_dataset.path('some-folder', 'iris2.csv')

            # With this we can open and read the file:
            with open(iris_file_path, 'r') as fp:
                iris_raw_data = fp.read()

            # The rest of your training script goes here

    """

    class DataSetNotFound(Exception):
        """[summary]"""

        pass


class AzureMLDataStore(MLDataStore):
    """[summary]"""

    @classmethod
    def from_model(
        cls,
        model: "Model",
        datasets: Iterable[Tuple[str, str]],
        workspace: Workspace,
        force_download: bool = False,
    ) -> "AzureMLDataStore":
        """[summary]

        Args:
            model (Model): [description]
            datasets (Iterable[Tuple[str, str]]): [description]
            workspace (Workspace): [description]
            force_download (bool, optional): [description]. Defaults to False.

        Returns:
            Description
        """
        mounted_datasets = {}

        for dataset_name, dataset_version in datasets:
            azureml_dataset = cls.load_azureml_dataset(
                workspace=workspace,
                dataset_name=dataset_name,
                dataset_version=dataset_version,
            )

            mounted_datasets[dataset_name] = cls.mount(
                model=model,
                azureml_dataset=azureml_dataset,
                force_download=force_download,
            )

        return cls(**mounted_datasets)

    @classmethod
    def load_azureml_dataset(
        cls, workspace: Workspace, dataset_name: str, dataset_version: str = None
    ) -> Union[TabularDataset, FileDataset]:
        """[summary]

        Args:
            workspace (Workspace): [description]
            dataset_name (str): [description]
            dataset_version (str, optional): [description]. Defaults to None.

        Raises:
            cls.DataSetNotFound: [description]

        Returns:
            Union[TabularDataset, FileDataset]: [description]
        """

        # azureml wants 'latest'
        if dataset_version is None:
            dataset_version = "latest"

        try:
            return Dataset.get_by_name(
                workspace=workspace, name=dataset_name, version=dataset_version
            )
        except exceptions._azureml_exception.UserErrorException:
            raise cls.DataSetNotFound(dataset_name)

    @classmethod
    def mount(
        cls,
        model: "Model",
        azureml_dataset: Union[TabularDataset, FileDataset],
        force_download: bool,
    ) -> MLDataSet:
        """[summary]

        Args:
            model (Model): [description]
            azureml_dataset (Union[TabularDataset, FileDataset]): [description]
            force_download (bool): [description]

        Raises:
            NotImplementedError: [description]

        Returns:
            MLDataSet: [description]
        """
        raise NotImplementedError


class MountedAzureMLDataStore(AzureMLDataStore):
    """[summary]"""

    @classmethod
    def mount(cls, model: "Model", azureml_dataset, force_download: bool) -> MLDataSet:
        """[summary]

        Args:
            model (Model): [description]
            azureml_dataset ([type]): [description]
            force_download (bool): [description]

        Returns:
            MLDataSet: [description]
        """
        # We don't need to download tabular datasets because we already know the schema.
        if isinstance(azureml_dataset, TabularDataset):
            df = azureml_dataset.to_pandas_dataframe()
            return MLDataSet(name=azureml_dataset.name, dataframe=df)

        else:
            mount_context = azureml_dataset.mount()
            mount_point = mount_context.mount_point
            mount_context.start()
            return MLDataSet(name=azureml_dataset.name, mount_path=mount_point)


class DownloadedAzureMLDataStore(AzureMLDataStore):
    """[summary]"""

    @classmethod
    def mount(
        cls, model: "Model", azureml_dataset: TabularDataset, force_download: bool
    ) -> MLDataSet:
        """[summary]

        Args:
            model (Model): [description]
            azureml_dataset (TabularDataset): [description]
            force_download (bool): [description]

        Returns:
            MLDataSet: [description]
        """

        # We don't need to download tabular datasets because we already know the schema.
        if isinstance(azureml_dataset, TabularDataset):
            df = azureml_dataset.to_pandas_dataframe()
            return MLDataSet(name=azureml_dataset.name, dataframe=df)

        mount_point = os.path.join(model.data_folder_path, azureml_dataset.name)
        try:
            azureml_dataset.download(mount_point, overwrite=force_download)
        except exceptions._azureml_exception.UserErrorException:
            # Dataset already exists on filesystem
            # TODO Rethink this solution
            print("NOTICE: Using cached dataset (from filesystem)")
        return MLDataSet(name=azureml_dataset.name, mount_path=mount_point)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

import click

from energinetml.core.backend import AbstractBackend
from energinetml.core.model import Model
from energinetml.settings import PACKAGE_NAME


def verify_compute_cluster(
    backend: "AbstractBackend",
    model: "Model",
    cluster_name: str,
    modify_vm_size: bool = False,
):
    """Verifies that the compute cluster exists.
    If not, the user is asked to choose another one.

    Args:
        backend (AbstractBackend): The backend to use.
        model (Model): Model which contains information about the compute cluster.
        cluster_name (str): The name of the compute cluster. This could be a new one.
        modify_vm_size (bool, optional): Boolean indicating if this method should
        change the vm_size. Defaults to False.

    Raises:
        click.Abort: If no compute clusters exist.
    """
    workspace = backend.get_workspace(model.project.as_dict())
    existing_clusters = backend.get_compute_clusters(workspace)
    existing_clusters_mapped = {c.name: c for c in existing_clusters}
    existing_cluster_names = [c.name for c in existing_clusters]

    if not existing_clusters:
        click.echo('No compute clusters exists in workspace "%s".' % workspace.name)
        click.echo('Run "%s cluster create" to create a new cluster.' % PACKAGE_NAME)
        raise click.Abort()

    while cluster_name not in existing_cluster_names:
        click.echo(
            (
                f"Could not find compute target '{cluster_name}' in the "
                f"available compute targets: {existing_cluster_names}."
            )
        )

        cluster_name = click.prompt(
            text="Please enter name of a compute cluster to use",
            type=click.Choice(existing_cluster_names),
        )

        click.echo(f"Using cluster '{cluster_name}' from now on.")

    cluster = existing_clusters_mapped[cluster_name]
    model.compute_target = cluster_name

    if modify_vm_size:
        model.vm_size = cluster.vm_size

    model.save()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

import os

import click

from energinetml.backend import default_backend as backend
from energinetml.cli.utils import (
    parse_input_path,
    parse_input_project_name,
    parse_input_resource_group,
    parse_input_subscription_id,
    parse_input_workspace_name,
)
from energinetml.core.project import MachineLearningProject
from energinetml.settings import COMMAND_NAME, DEFAULT_LOCATION

PROJECT_FILES = ("project.json", "requirements.txt")


# -- CLI Command -------------------------------------------------------------


@click.command()
@click.option(
    "--path",
    "-p",
    default=None,
    type=click.Path(dir_okay=True, resolve_path=True),
    callback=parse_input_path(PROJECT_FILES),
    help="Project path (default to current)",
)
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    type=str,
    callback=parse_input_project_name(),
    help="Project name",
)
@click.option(
    "--subscription",
    "-s",
    "subscription_id",
    required=False,
    default=None,
    type=str,
    callback=parse_input_subscription_id(),
    help="Azure subscription name",
)
@click.option(
    "--resource-group",
    "-r",
    "resource_group",
    required=False,
    default=None,
    type=str,
    callback=parse_input_resource_group(),
    help="Azure Resource Group",
)
@click.option(
    "--workspace",
    "-w",
    "workspace_name",
    required=False,
    default=None,
    type=str,
    callback=parse_input_workspace_name(),
    help="AzureML Workspace name",
)
@click.option(
    "--location",
    "-l",
    "location",
    default=DEFAULT_LOCATION,
    required=False,
    type=str,
    help=f"Azure location (default: {DEFAULT_LOCATION})",
)
def init_project(
    path: str,
    name: str,
    subscription_id: str,
    resource_group: str,
    workspace_name: str,
    location: str,
):
    """Create a new, empty machine learning project."""
    if not os.path.isdir(path):
        os.makedirs(path)

    project_meta = {
        "subscription_id": subscription_id,
        "resource_group": resource_group,
        "workspace_name": workspace_name,
    }
    workspace = backend.get_workspace(project_meta)

    MachineLearningProject.create(
        path=path,
        name=name,
        subscription_id=subscription_id,
        resource_group=resource_group,
        workspace_name=workspace_name,
        location=location,
        vnet_name=workspace.tags["vnet_name"],
        subnet_name=workspace.tags["subnet_name"],
    )

    click.echo("-" * 79)
    click.echo(f"Initialized the project at: {path}")
    click.echo(f"Type '{COMMAND_NAME} model init' to add a new model to the project.")
    click.echo("-" * 79)



from unittest.mock import PropertyMock, patch

from click.testing import CliRunner

from energinetml.cli.model.files import files
from energinetml.core.model import Model
from tests.constants import FILES_INCLUDE

# -- create() Tests ----------------------------------------------------------


@patch("energinetml.cli.model.files.click.echo")
@patch.object(Model, "files", new_callable=PropertyMock)
def test__model_files__should_echo_model_files(model_files_mock, echo_mock, model):
    """
    :param Mock model_files_mock:
    :param Mock echo_mock:
    """
    runner = CliRunner()

    model_files_mock.return_value = FILES_INCLUDE

    # Act
    result = runner.invoke(cli=files, args=["--path", model.path])

    # Assert
    assert result.exit_code == 0
    assert echo_mock.call_count == len(FILES_INCLUDE)

    for file_name in FILES_INCLUDE:
        echo_mock.assert_any_call(file_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import click

from energinetml.cli.utils import discover_model
from energinetml.core.model import Model


@click.command()
@discover_model()
def files(model: Model) -> None:
    """List files that are copied to the cloud when submitting."""
    for file_path in sorted(model.files):
        click.echo(file_path)



"""[summary]
"""
import sys
from datetime import datetime

import pytest

from energinetml.core.logger import ConsoleLogger, MetricsLogger
from energinetml.settings import DEFAULT_LOG_ENCODING


@pytest.fixture()
def clog(monkeypatch, tmp_path):
    monkeypatch.setattr("energinetml.core.logger.DEFAULT_LOG_PATH", tmp_path)
    yield ConsoleLogger(name="stdout", console=sys.stdout)


LOG_TEXT = "This messages come from the unit test.\n"
N_WRITES = 5


class TestConsoleLogger:
    """Unittests for the ConsoleLogger class."""

    def test__isatty(self, clog):
        """
        Test for requried class function.
        This function must return flase.
        """
        assert not clog.isatty()

    def test__flush(self, clog):
        """Test for requried class function."""
        clog.flush()

    def test__write(self, clog):
        """Test for requried class function."""

        # Act
        for _ in range(N_WRITES):
            clog.write(f"{datetime.now()} {LOG_TEXT}")

        clog.flush()

        # Assert
        with open(clog.filepath, "r", encoding=DEFAULT_LOG_ENCODING) as handle:
            lines = handle.readlines()

        assert len(lines) == N_WRITES
        for text in lines:
            assert text.endswith(LOG_TEXT)


@pytest.fixture
def logger():
    yield MetricsLogger()


class TestMetricsLogger:
    def test__echo(self, logger):
        """
        :param MetricsLogger logger:
        """
        with pytest.raises(NotImplementedError):
            logger.echo("s")

    def test__log(self, logger):
        """
        :param MetricsLogger logger:
        """
        with pytest.raises(NotImplementedError):
            logger.log("name", "value")

    def test__tag(self, logger):
        """
        :param MetricsLogger logger:
        """
        with pytest.raises(NotImplementedError):
            logger.tag("key", "value")

    def test__dataframe(self, logger):
        """
        :param MetricsLogger logger:
        """
        with pytest.raises(NotImplementedError):
            logger.dataframe("name", "df")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import os
import re
from functools import cached_property
from typing import TYPE_CHECKING, Dict, List

import azureml
from azure.mgmt.resource import ResourceManagementClient, SubscriptionClient
from azureml._common.exceptions import AzureMLException
from azureml.core import ComputeTarget, Environment, Experiment
from azureml.core import Model as AzureMLModel
from azureml.core import Run, RunConfiguration, ScriptRunConfig, Workspace
from azureml.core.authentication import InteractiveLoginAuthentication
from azureml.core.compute import AmlCompute
from azureml.core.conda_dependencies import CondaDependencies

from energinetml.core.backend import AbstractBackend
from energinetml.core.project import Project
from energinetml.settings import (
    CLUSTER_IDLE_SECONDS_BEFORE_SCALEDOWN,
    PACKAGE_NAME,
    PYTHON_VERSION,
)

from .submitting import AzureSubmitContext
from .training import AzureCloudTrainingContext, AzureLocalTrainingContext

if TYPE_CHECKING:
    from energinetml.core.model import Model as EnerginetMLModel


class AzureBackend(AbstractBackend):
    def parse_azureml_exception(
        self, e: AzureMLException
    ) -> AbstractBackend.BackendException:
        """Extracts error message from AzureMLException and
        returns a BackendException.

        Args:
            e (AzureMLException): [description]

        Returns:
            AbstractBackend.BackendException: [description]
        """
        msg = str(e)
        matches = re.findall(r'"message":\s*"([^"]+)"', msg)
        if matches:
            return self.BackendException(matches[0])
        else:
            return self.BackendException(msg)

    @cached_property
    def _credential(self) -> InteractiveLoginAuthentication:
        """[summary]

        Returns:
            InteractiveLoginAuthentication: [description]
        """
        return InteractiveLoginAuthentication()

    def get_available_subscriptions(self) -> List[str]:
        """[summary]

        Raises:
            self.parse_azureml_exception: [description]

        Returns:
            List[str]: [description]
        """
        subscription_client = SubscriptionClient(self._credential)
        try:
            return list(subscription_client.subscriptions.list())
        except azureml._common.exceptions.AzureMLException as e:
            raise self.parse_azureml_exception(e)

    def get_available_resource_groups(self, subscription_id: str) -> List[str]:
        """[summary]

        Args:
            subscription_id (str): [description]

        Raises:
            self.parse_azureml_exception: [description]

        Returns:
            List[str]: [description]
        """
        resource_client = ResourceManagementClient(self._credential, subscription_id)
        try:
            return list(resource_client.resource_groups.list())
        except azureml._common.exceptions.AzureMLException as e:
            raise self.parse_azureml_exception(e)

    # -- Workspaces ----------------------------------------------------------

    def get_available_workspaces(
        self, subscription_id: str, resource_group: str
    ) -> List[Workspace]:
        """[summary]

        Args:
            subscription_id (str): [description]
            resource_group (str): [description]

        Raises:
            self.parse_azureml_exception: [description]

        Returns:
            List[Workspace]: [description]
        """
        try:
            workspaces_mapped = Workspace.list(
                auth=self._credential,
                subscription_id=subscription_id,
                resource_group=resource_group,
            )
        except azureml._common.exceptions.AzureMLException as e:
            raise self.parse_azureml_exception(e)

        workspaces = []

        for workspace_list in workspaces_mapped.values():
            workspaces.extend(workspace_list)

        return workspaces

    def get_available_workspace_names(
        self, subscription_id: str, resource_group: str
    ) -> List[str]:
        """[summary]

        Args:
            subscription_id (str): [description]
            resource_group (str): [description]

        Returns:
            List[str]: [description]
        """
        available_workspaces = self.get_available_workspaces(
            subscription_id=subscription_id, resource_group=resource_group
        )

        return [w.name for w in available_workspaces]

    def get_workspace(self, project_meta: Dict[str, str]) -> Workspace:
        """_summary_

        Args:
            project_meta (dict): _description_

        Raises:
            self.parse_azureml_exception: _description_

        Returns:
            Workspace: _description_
        """

        expected_keywords = ["subscription_id", "resource_group", "workspace_name"]
        for keyword in expected_keywords:
            if keyword not in project_meta:
                raise Project.ConfigError(f"Expected '{keyword}' as a key.")

        try:
            return Workspace.get(
                auth=self._credential,
                subscription_id=project_meta["subscription_id"],
                resource_group=project_meta["resource_group"],
                name=project_meta["workspace_name"],
            )
        except azureml._common.exceptions.AzureMLException as e:
            raise self.parse_azureml_exception(e)

    # -- Compute clusters ----------------------------------------------------

    def get_compute_clusters(self, workspace: Workspace) -> List[AmlCompute]:
        """[summary]

        Args:
            workspace (Workspace): [description]

        Raises:
            self.parse_azureml_exception: [description]

        Returns:
            List[AmlCompute]: [description]
        """
        try:
            return AmlCompute.list(workspace=workspace)
        except azureml._common.exceptions.AzureMLException as e:
            raise self.parse_azureml_exception(e)

    def get_available_vm_sizes(self, workspace: Workspace) -> List[str]:
        """[summary]

        Args:
            workspace (Workspace): [description]

        Raises:
            self.parse_azureml_exception: [description]

        Returns:
            List[str]: [description]
        """
        try:
            return AmlCompute.supported_vmsizes(workspace=workspace)
        except azureml._common.exceptions.AzureMLException as e:
            raise self.parse_azureml_exception(e)

    def create_compute_cluster(
        self,
        workspace: Workspace,
        name: str,
        vm_size: str,
        min_nodes: int,
        max_nodes: int,
        vnet_resource_group_name: str,
        vnet_name: str,
        subnet_name: str,
    ) -> None:
        """[summary]

        Args:
            workspace (Workspace): [description]
            name (str): [description]
            vm_size (str): [description]
            min_nodes (int): [description]
            max_nodes (int): [description]
            vnet_resource_group_name (str): [description]
            vnet_name (str): [description]
            subnet_name (str): [description]

        Raises:
            self.parse_azureml_exception: [description]
        """

        compute_config = AmlCompute.provisioning_configuration(
            vm_size=vm_size,
            min_nodes=min_nodes,
            max_nodes=max_nodes,
            vnet_resourcegroup_name=vnet_resource_group_name,
            vnet_name=vnet_name,
            subnet_name=subnet_name,
            idle_seconds_before_scaledown=CLUSTER_IDLE_SECONDS_BEFORE_SCALEDOWN,
        )

        try:
            ComputeTarget.create(workspace, name, compute_config).wait_for_completion(
                show_output=False
            )
        except azureml._common.exceptions.AzureMLException as e:
            raise self.parse_azureml_exception(e)

    # -- Contexts ------------------------------------------------------------

    def get_local_training_context(
        self, force_download: bool
    ) -> AzureLocalTrainingContext:
        """[summary]

        Args:
            force_download (bool): [description]

        Returns:
            AzureLocalTrainingContext: [description]
        """
        return AzureLocalTrainingContext(self, force_download)

    def get_cloud_training_context(self) -> AzureCloudTrainingContext:
        """[summary]

        Returns:
            AzureCloudTrainingContext: [description]
        """
        return AzureCloudTrainingContext()

    def submit_model(
        self, model: "EnerginetMLModel", params: List[str]
    ) -> AzureSubmitContext:
        """[summary]

        Args:
            model (Model): [description]
            params (List[str]): [description]

        Raises:
            self.parse_azureml_exception: [description]

        Returns:
            AzureSubmitContext: [description]
        """
        cd = CondaDependencies()
        cd.set_python_version(PYTHON_VERSION)

        # Project requirements (from requirements.txt)
        for requirement in model.requirements:
            cd.add_pip_package(requirement.line)

        # Python environment
        env = Environment(model.requirements.get(PACKAGE_NAME).line)
        env.python.conda_dependencies = cd

        compute_config = RunConfiguration()
        compute_config.target = model.compute_target
        compute_config.environment = env

        project_meta = model.project.as_dict()
        workspace = self.get_workspace(project_meta)

        experiment = Experiment(workspace=workspace, name=model.experiment)

        # Change the python module path to a folder path,
        # so that we can cd to and run the script in the cloud.
        snapshot_relative_path = model.module_name.replace(".", os.path.sep)

        with model.temporary_folder() as path:
            config = ScriptRunConfig(
                source_directory=path,
                # Command to enter the module directory and running a cloud training
                # The 'python .'-command runs the energinetml-cli.
                command=[
                    "cd",
                    snapshot_relative_path,
                    "&&",
                    "python",
                    ".",
                    "model",
                    "train",
                    "--cloud-mode",
                ]
                + list(params),
                run_config=compute_config,
            )
            try:
                run = experiment.submit(config)
            except azureml._common.exceptions.AzureMLException as e:
                raise self.parse_azureml_exception(e)

            return AzureSubmitContext(model, run)

    def release_model(
        self, workspace: Workspace, model_path: str, model_name: str, **kwargs
    ) -> AzureMLModel:
        """[summary]

        Args:
            workspace (Workspace): [description]
            model_path (str): [description]
            model_name (str): [description]

        Raises:
            self.parse_azureml_exception: [description]

        Returns:
            AzureMLModel: [description]
        """
        try:
            asset = AzureMLModel._create_asset(
                workspace.service_context, model_path, model_name, None
            )

            return AzureMLModel._register_with_asset(
                workspace=workspace, model_name=model_name, asset_id=asset.id, **kwargs
            )
        except azureml._common.exceptions.AzureMLException as e:
            raise self.parse_azureml_exception(e)

    def download_model_files(
        self, workspace: Workspace, experiment_name: str, run_id: str, folder_path: str
    ) -> str:
        """Downloads model files to folder including snapshot (model.py, model.json, etc) as a
        zip-file and trainedmodel filesfrom the Azure ML cloud,
        based on an `experiment_name` and a `run_id`.

        Args:
            workspace (Workspace): An Azure workspace
            experiment_name (str): The name of the experiment
            run_id (str): The id of the run

        Returns:
        str: The path to the snapshot zip-file.
        """

        experiment = Experiment(workspace, experiment_name)
        run = Run(experiment, run_id)

        # Download project snapshot (model.py, model.json, etc).
        # Snapshot is downloaded as a zip-archive.
        print("Downloading model code snapshot")
        snapshot_zip_path = run.restore_snapshot(path=folder_path)

        print("Downloading trained model file")
        run.download_files(output_directory=folder_path)

        return snapshot_zip_path

    def get_portal_url(self, project_meta: dict, experiment_name: str, run_id: str):
        """Get the url to the Azure ML portal for a specific run

        Args:
            project_meta (dict): Project information for retrieving workspace
            experiment_name (str): Name of the experiment
            run_id (str): Id of the run
        """
        workspace = self.get_workspace(project_meta)
        experiment = Experiment(workspace, experiment_name)
        run = Run(experiment, run_id)

        return run.get_portal_url()
